import shutil
from pathlib import Path
from uuid import uuid4

import pytest

from panimg.models import ImageType, PanImgFile
from panimg.post_processors.tiff_to_dzi import tiff_to_dzi
from tests import RESOURCE_PATH


def test_dzi_creation(tmpdir_factory):
    filename = "valid_tiff.tif"

    temp_file = Path(tmpdir_factory.mktemp("temp") / filename)
    shutil.copy(RESOURCE_PATH / filename, temp_file)

    image_file = PanImgFile(
        image_id=uuid4(), image_type=ImageType.TIFF, file=temp_file
    )

    result = tiff_to_dzi(image_files={image_file})

    assert len(result.new_image_files) == 1

    new_file = result.new_image_files.pop()

    assert new_file.image_id == image_file.image_id
    assert new_file.image_type == ImageType.DZI
    assert (
        new_file.file == image_file.file.parent / f"{image_file.image_id}.dzi"
    )
    assert (
        new_file.directory
        == image_file.file.parent / f"{image_file.image_id}_files"
    )

    assert len(list((new_file.directory).rglob("*.jpeg"))) == 9


def test_no_exception_when_failed(tmpdir_factory, caplog):
    filename = "no_dzi.tif"

    temp_file = Path(tmpdir_factory.mktemp("temp") / filename)
    shutil.copy(RESOURCE_PATH / filename, temp_file)

    image_file = PanImgFile(
        image_id=uuid4(), image_type=ImageType.TIFF, file=temp_file
    )

    result = tiff_to_dzi(image_files={image_file})

    assert len(result.new_image_files) == 0

    # The last warning should be from our logger
    last_log = caplog.records[-1]
    assert last_log.name == "panimg.post_processors.tiff_to_dzi"
    assert last_log.levelname == "WARNING"
    assert "Could not create DZI for" in last_log.message


@pytest.mark.parametrize("image_type", (ImageType.DZI, ImageType.MHD))
def test_non_tiff_skipped(tmpdir_factory, image_type):
    filename = "valid_tiff.tif"

    temp_file = Path(tmpdir_factory.mktemp("temp") / filename)
    shutil.copy(RESOURCE_PATH / filename, temp_file)

    image_file = PanImgFile(
        image_id=uuid4(), image_type=image_type, file=temp_file
    )

    result = tiff_to_dzi(image_files={image_file})

    assert len(result.new_image_files) == 0



import logging

from panimg.models import ImageType, PanImgFile, PostProcessorResult
from panimg.settings import DZI_TILE_SIZE

logger = logging.getLogger(__name__)


def tiff_to_dzi(*, image_files: set[PanImgFile]) -> PostProcessorResult:
    new_image_files: set[PanImgFile] = set()

    for file in image_files:
        if file.image_type == ImageType.TIFF:
            try:
                result = _create_dzi_image(tiff_file=file)
            except Exception:
                logger.warning(f"Could not create DZI for {file}")
                continue

            new_image_files |= result.new_image_files

    return PostProcessorResult(new_image_files=new_image_files)


def _create_dzi_image(*, tiff_file: PanImgFile) -> PostProcessorResult:
    import pyvips

    # Creates a dzi file and corresponding tiles in directory {pk}_files
    dzi_output = tiff_file.file.parent / str(tiff_file.image_id)

    image = pyvips.Image.new_from_file(
        str(tiff_file.file.absolute()), access="sequential"
    )

    pyvips.Image.dzsave(image, str(dzi_output), tile_size=DZI_TILE_SIZE)

    new_file = PanImgFile(
        image_id=tiff_file.image_id,
        image_type=ImageType.DZI,
        file=(dzi_output.parent / f"{dzi_output.name}.dzi").absolute(),
        directory=(dzi_output.parent / f"{dzi_output.name}_files").absolute(),
    )

    return PostProcessorResult(new_image_files={new_file})



import os
import shutil
from pathlib import Path
from uuid import uuid4

import pytest
import tifffile as tiff_lib
from pytest import approx
from tifffile import tifffile

from panimg.exceptions import ValidationError
from panimg.image_builders.tiff import (
    GrandChallengeTiffFile,
    _extract_tags,
    _get_color_space,
    _load_with_openslide,
    _load_with_tiff,
    image_builder_tiff,
)
from panimg.models import MAXIMUM_SEGMENTS_LENGTH, ColorSpace
from panimg.panimg import _build_files
from tests import RESOURCE_PATH


@pytest.mark.parametrize(
    "color_space_string, expected",
    [
        ("GRAY", ColorSpace.GRAY),
        ("MINISBLACK", ColorSpace.GRAY),
        ("minisblack", ColorSpace.GRAY),
        ("RGB", ColorSpace.RGB),
        ("RGBA", ColorSpace.RGBA),
        ("YCBCR", ColorSpace.YCBCR),
        ("Colour", None),
    ],
)
def test_get_color_space(color_space_string, expected):
    color_space = None

    try:
        color_space = _get_color_space(color_space_string=color_space_string)
    except ValidationError:
        pass

    assert color_space is expected


@pytest.mark.parametrize(
    (
        "path,"
        "color_space,"
        "resolution_levels,"
        "image_height,"
        "image_width,"
        "voxel_height_mm,"
        "voxel_width_mm,"
        "expected_error_message"
    ),
    [
        ("dummy.tiff", 1, 1, 10, 10, 0.1, 0.1, ""),
        (
            "dummy.tiff",
            1,
            None,
            10,
            10,
            0.1,
            0.1,
            "Not a valid tif: Resolution levels not valid",
        ),
        (
            "dummy.tiff",
            1,
            1,
            None,
            10,
            0.1,
            0.1,
            "Not a valid tif: Image height could not be determined",
        ),
        (
            "dummy.tiff",
            1,
            1,
            10,
            None,
            0.1,
            0.1,
            "Not a valid tif: Image width could not be determined",
        ),
        (
            "dummy.tiff",
            1,
            1,
            10,
            10,
            None,
            0.1,
            "Not a valid tif: Voxel height could not be determined",
        ),
        (
            "dummy.tiff",
            1,
            1,
            10,
            10,
            0.1,
            None,
            "Not a valid tif: Voxel width could not be determined",
        ),
    ],
)
def test_grandchallengetifffile_validation(
    path,
    color_space,
    resolution_levels,
    image_height,
    image_width,
    voxel_height_mm,
    voxel_width_mm,
    expected_error_message,
):
    error_message = ""

    try:
        gc_file = GrandChallengeTiffFile(Path(path))
        gc_file.color_space = color_space
        gc_file.resolution_levels = resolution_levels
        gc_file.image_height = image_height
        gc_file.image_width = image_width
        gc_file.voxel_height_mm = voxel_height_mm
        gc_file.voxel_width_mm = voxel_width_mm
        gc_file.validate()
    except ValidationError as e:
        error_message = str(e)

    assert expected_error_message in error_message
    if not expected_error_message:
        assert not error_message


@pytest.mark.parametrize(
    "source_dir, filename, error_message, min, max, segments",
    [
        (RESOURCE_PATH, "test_min_max.tif", "", 0, 4, {0, 1, 2, 3, 4}),
        (
            RESOURCE_PATH,
            "invalid_resolutions_tiff.tif",
            "Invalid resolution unit NONE in tiff file",
            None,
            None,
            {},
        ),
    ],
)
def test_load_with_tiff(
    source_dir,
    filename,
    error_message,
    min,
    max,
    segments,
    tmpdir_factory,
):
    error = ""
    # Copy resource file to writable temp directory
    temp_file = Path(tmpdir_factory.mktemp("temp") / filename)
    shutil.copy(source_dir / filename, temp_file)
    gc_file = GrandChallengeTiffFile(temp_file)
    gc_file.pk = uuid4()
    try:
        gc_file = _load_with_tiff(gc_file=gc_file)
        assert gc_file.min_voxel_value == min
        assert gc_file.max_voxel_value == max
        assert gc_file.segments == segments
    except ValidationError as e:
        error = str(e)

    assert error_message in error
    if not error_message:
        assert not error


def test_segments_computed_property():
    gc_file = GrandChallengeTiffFile(RESOURCE_PATH / "test_min_max.tif")
    assert gc_file.segments is None
    gc_file.min_voxel_value = 0
    gc_file.max_voxel_value = MAXIMUM_SEGMENTS_LENGTH
    assert gc_file.segments is None
    gc_file.max_voxel_value = 4
    assert gc_file.segments == frozenset({0, 1, 2, 3, 4})


@pytest.mark.parametrize(
    "source_dir, filename",
    [(RESOURCE_PATH, "valid_tiff.tif"), (RESOURCE_PATH, "no_dzi.tif")],
)
def test_load_with_open_slide(source_dir, filename, tmpdir_factory):
    # Copy resource file to writable temp directory
    temp_file = Path(tmpdir_factory.mktemp("temp") / filename)
    shutil.copy(source_dir / filename, temp_file)
    gc_file = GrandChallengeTiffFile(temp_file)

    output_dir = Path(tmpdir_factory.mktemp("output"))
    (output_dir / filename).mkdir()
    gc_file = _load_with_tiff(gc_file=gc_file)
    gc_file = _load_with_openslide(gc_file=gc_file)

    assert gc_file.validate() is None


@pytest.mark.parametrize(
    "resource, expected_error_message, voxel_size",
    [
        (
            RESOURCE_PATH / "valid_tiff.tif",
            "",
            [1, 1],
        )
    ],
)
def test_tiff_image_entry_creation(
    resource, expected_error_message, voxel_size
):
    error_message = ""
    gc_file = GrandChallengeTiffFile(resource)
    try:
        tiff_file = tifffile.TiffFile(str(gc_file.path.absolute()))
        gc_file = _extract_tags(
            gc_file=gc_file,
            pages=tiff_file.pages,
            byteorder=tiff_file.byteorder,
        )
    except ValidationError as e:
        error_message = str(e)

    # Asserts possible file opening failures
    assert expected_error_message in error_message
    if not expected_error_message:
        assert not error_message

    # Asserts successful creation data
    if not expected_error_message:
        tiff_file = tiff_lib.TiffFile(str(resource.absolute()))
        tiff_tags = tiff_file.pages[0].tags

        assert gc_file.path.name == resource.name
        assert gc_file.image_width == tiff_tags["ImageWidth"].value
        assert gc_file.image_height == tiff_tags["ImageLength"].value
        assert gc_file.resolution_levels == len(tiff_file.pages)
        assert gc_file.color_space == _get_color_space(
            color_space_string=tiff_tags[
                "PhotometricInterpretation"
            ].value.name
        )
        assert gc_file.voxel_width_mm == approx(voxel_size[0])
        assert gc_file.voxel_height_mm == approx(voxel_size[1])
        assert gc_file.min_voxel_value is None
        assert gc_file.max_voxel_value is None
        assert gc_file.segments is None


# Integration test of all features being accessed through the image builder
def test_image_builder_tiff(tmpdir_factory):
    # Copy resource files to writable temp directory
    temp_dir = Path(tmpdir_factory.mktemp("temp") / "resources")
    output_dir = Path(tmpdir_factory.mktemp("output"))

    shutil.copytree(
        RESOURCE_PATH,
        temp_dir,
        ignore=shutil.ignore_patterns(
            "dicom*", "complex_tiff", "dzi_tiff", "isyntax_wsi"
        ),
    )
    files = [Path(d[0]).joinpath(f) for d in os.walk(temp_dir) for f in d[2]]

    image_builder_result = _build_files(
        builder=image_builder_tiff, files=files, output_directory=output_dir
    )

    expected_files = [
        temp_dir / "valid_tiff.tif",
        temp_dir / "no_dzi.tif",
        temp_dir / "test_min_max.tif",
    ]

    assert sorted(image_builder_result.consumed_files) == sorted(
        expected_files
    )

    file_to_pk = {i.name: i.pk for i in image_builder_result.new_images}

    for file in expected_files:
        pk = file_to_pk[file.name]
        assert os.path.isfile(output_dir / file.name / f"{pk}.tif")

    # Assert that both tiff images are imported
    assert len(image_builder_result.new_image_files) == 3


@pytest.mark.xfail(
    reason="skip for now as we don't want to upload a large testset"
)
@pytest.mark.parametrize(
    "resource",
    [
        RESOURCE_PATH / "convert_to_tiff" / "vms",
        RESOURCE_PATH / "convert_to_tiff" / "svs",
        RESOURCE_PATH / "convert_to_tiff" / "ndpi",
        RESOURCE_PATH / "convert_to_tiff" / "scn",
        RESOURCE_PATH / "convert_to_tiff" / "mrxs",
        RESOURCE_PATH / "convert_to_tiff" / "bif",
        RESOURCE_PATH / "convert_to_tiff" / "isyntax",
    ],
)
def test_convert_to_tiff(resource, tmpdir_factory):
    output_dir = Path(tmpdir_factory.mktemp("output"))

    input_files = {f for f in resource.glob("*") if f.is_file()}

    result = _build_files(
        builder=image_builder_tiff,
        files=input_files,
        output_directory=output_dir,
    )

    assert len(result.new_images) == 1
    assert len(result.new_image_files) == 1


def test_error_handling(tmpdir_factory):
    # Copy resource files to writable temp directory
    # The content files are dummy files and won't compile to tiff.
    # The point is to test the loading of gc_files and make sure all
    # related files are associated with the gc_file
    temp_dir = Path(tmpdir_factory.mktemp("temp") / "resources")
    shutil.copytree(RESOURCE_PATH / "complex_tiff", temp_dir)
    files = {Path(d[0]).joinpath(f) for d in os.walk(temp_dir) for f in d[2]}

    image_builder_result = _build_files(
        builder=image_builder_tiff,
        files=files,
        output_directory=Path(tmpdir_factory.mktemp("output")),
    )

    output = {k.name: v for k, v in image_builder_result.file_errors.items()}

    assert output == {
        "Mirax2-Fluorescence-1.mrxs": [
            "TIFF image builder: Could not convert file to TIFF: "
            "Mirax2-Fluorescence-1.mrxs",
            "TIFF image builder: Could not open file with tifffile.",
            "TIFF image builder: Could not open file with OpenSlide.",
            "TIFF image builder: Validation error: "
            "Not a valid tif: Image width could not be determined.",
        ],
        "CMU-1-40x - 2010-01-12 13.24.05.vms": [
            "TIFF image builder: Could not convert file to TIFF: "
            "CMU-1-40x - 2010-01-12 13.24.05.vms",
            "TIFF image builder: Could not open file with tifffile.",
            "TIFF image builder: Could not open file with OpenSlide.",
            "TIFF image builder: Validation error: "
            "Not a valid tif: Image width could not be determined.",
        ],
        "Data0001.dat": [
            "TIFF image builder: Could not open file with tifffile.",
            "TIFF image builder: Could not open file with OpenSlide.",
            "TIFF image builder: Validation error: "
            "Not a valid tif: Image width could not be determined.",
        ],
        "Index.dat": [
            "TIFF image builder: Could not open file with tifffile.",
            "TIFF image builder: Could not open file with OpenSlide.",
            "TIFF image builder: Validation error: "
            "Not a valid tif: Image width could not be determined.",
        ],
        "CMU-1-40x - 2010-01-12 13.24.05(0,1).jpg": [
            "TIFF image builder: Could not open file with tifffile.",
            "TIFF image builder: Validation error: "
            "Not a valid tif: Image width could not be determined.",
        ],
        "CMU-1-40x - 2010-01-12 13.24.05_map2.jpg": [
            "TIFF image builder: Could not open file with tifffile.",
            "TIFF image builder: Validation error: "
            "Not a valid tif: Image width could not be determined.",
        ],
        "Slidedat.ini": [
            "TIFF image builder: Could not open file with tifffile.",
            "TIFF image builder: Could not open file with OpenSlide.",
            "TIFF image builder: Validation error: "
            "Not a valid tif: Image width could not be determined.",
        ],
        "CMU-1-40x - 2010-01-12 13.24.05_macro.jpg": [
            "TIFF image builder: Could not open file with tifffile.",
            "TIFF image builder: Validation error: "
            "Not a valid tif: Image width could not be determined.",
        ],
        "CMU-1-40x - 2010-01-12 13.24.05.opt": [
            "TIFF image builder: Could not open file with tifffile.",
            "TIFF image builder: Validation error: "
            "Not a valid tif: Image width could not be determined.",
        ],
        "CMU-1-40x - 2010-01-12 13.24.05(1,0).jpg": [
            "TIFF image builder: Could not open file with tifffile.",
            "TIFF image builder: Validation error: "
            "Not a valid tif: Image width could not be determined.",
        ],
        "Data0000.dat": [
            "TIFF image builder: Could not open file with tifffile.",
            "TIFF image builder: Could not open file with OpenSlide.",
            "TIFF image builder: Validation error: "
            "Not a valid tif: Image width could not be determined.",
        ],
        "CMU-1-40x - 2010-01-12 13.24.05.jpg": [
            "TIFF image builder: Could not open file with tifffile.",
            "TIFF image builder: Validation error: "
            "Not a valid tif: Image width could not be determined.",
        ],
        "CMU-1-40x - 2010-01-12 13.24.05(1,1).jpg": [
            "TIFF image builder: Could not open file with tifffile.",
            "TIFF image builder: Validation error: "
            "Not a valid tif: Image width could not be determined.",
        ],
        "Data0002.dat": [
            "TIFF image builder: Could not open file with tifffile.",
            "TIFF image builder: Could not open file with OpenSlide.",
            "TIFF image builder: Validation error: "
            "Not a valid tif: Image width could not be determined.",
        ],
    }



import shutil
from pathlib import Path
from uuid import uuid4

import pytest

from panimg import convert, post_process
from panimg.models import ImageType, PanImgFile, PostProcessorResult
from tests import RESOURCE_PATH


@pytest.mark.parametrize("post_processors", (None, []))
def test_dzi_creation(tmpdir_factory, post_processors):
    filenames = [
        "valid_tiff.tif",
        "no_dzi.tif",
        "invalid_resolutions_tiff.tif",
    ]

    input_dir = Path(tmpdir_factory.mktemp("input"))
    output_dir = Path(tmpdir_factory.mktemp("output"))

    for f in filenames:
        shutil.copy(RESOURCE_PATH / f, input_dir / f)

    result = convert(
        input_directory=input_dir,
        output_directory=output_dir,
        post_processors=post_processors,
    )

    assert len(result.new_images) == 2

    if post_processors is None:
        assert len(result.new_image_files) == 3
        assert (
            len([f for f in result.new_image_files if f.directory is None])
            == 2
        )
    else:
        assert len(result.new_image_files) == 2
        assert (
            len([f for f in result.new_image_files if f.directory is None])
            == 2
        )


def bad_post_processor(*, image_files: set[PanImgFile]) -> PostProcessorResult:
    good_files = {
        PanImgFile(
            image_id=f.image_id, image_type=f.image_type, file=Path("foo")
        )
        for f in image_files
    }
    bad_files = {
        PanImgFile(image_id=uuid4(), image_type=f.image_type, file=Path("foo"))
        for f in image_files
    }

    good_directories = {
        PanImgFile(
            image_id=f.image_id,
            image_type=f.image_type,
            file=Path("foo"),
            directory=Path("foo_files"),
        )
        for f in image_files
    }
    bad_directories = {
        PanImgFile(
            image_id=uuid4(),
            image_type=f.image_type,
            file=Path("foo"),
            directory=Path("foo_files"),
        )
        for f in image_files
    }

    return PostProcessorResult(
        new_image_files=good_files
        | bad_files
        | good_directories
        | bad_directories,
    )


def test_post_processors_are_filtered():
    image_files = {
        PanImgFile(image_id=uuid4(), image_type=t, file=f"{t}")
        for t in ImageType
    }
    existing_ids = {f.image_id for f in image_files}

    raw_result = bad_post_processor(image_files=image_files)

    # The bad processor should produce twice as many outputs than inputs
    assert len(image_files) == 3
    assert len(raw_result.new_image_files) == 12

    # The bad results should be filtered out
    result = post_process(
        image_files=image_files, post_processors=[bad_post_processor]
    )

    assert len(result.new_image_files) == 6
    assert len([f for f in result.new_image_files if f.directory is None]) == 3
    assert {f.image_id for f in result.new_image_files} == existing_ids



import os
import re
from collections import defaultdict
from collections.abc import Callable, Iterator
from dataclasses import dataclass, field
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import DefaultDict
from uuid import UUID, uuid4

import tifffile

from panimg.contrib.wsi_dcm_to_tiff.dcm_to_tiff import (
    dcm_to_tiff as wsi_dcm_to_tiff,
)
from panimg.contrib.wsi_isyntax_to_tiff.isyntax_to_tiff import (
    isyntax_to_tiff as wsi_isyntax_to_tiff,
)
from panimg.exceptions import UnconsumedFilesException, ValidationError
from panimg.image_builders.dicom import get_dicom_headers_by_study
from panimg.models import MAXIMUM_SEGMENTS_LENGTH, ColorSpace, TIFFImage

try:
    from isyntax import ISyntax
except ImportError:
    _has_isyntax = False
else:
    _has_isyntax = True

DICOM_WSI_STORAGE_ID = "1.2.840.10008.5.1.4.1.1.77.1.6"


def format_error(message: str) -> str:
    return f"TIFF image builder: {message}"


@dataclass
class GrandChallengeTiffFile:
    path: Path
    pk: UUID = field(default_factory=uuid4)
    image_width: int = 0
    image_height: int = 0
    resolution_levels: int = 0
    color_space: ColorSpace | None = None
    voxel_width_mm: float = 0
    voxel_height_mm: float = 0
    associated_files: list[Path] = field(default_factory=list)
    min_voxel_value: int | None = None
    max_voxel_value: int | None = None

    def validate(self) -> None:
        if not self.image_width:
            raise ValidationError(
                "Not a valid tif: Image width could not be determined"
            )
        if not self.image_height:
            raise ValidationError(
                "Not a valid tif: Image height could not be determined"
            )
        if not self.resolution_levels:
            raise ValidationError(
                "Not a valid tif: Resolution levels not valid"
            )
        if not self.voxel_width_mm:
            raise ValidationError(
                "Not a valid tif: Voxel width could not be determined"
            )
        if not self.voxel_height_mm:
            raise ValidationError(
                "Not a valid tif: Voxel height could not be determined"
            )
        if not self.color_space:
            raise ValidationError(
                "Not a valid tif: Color space could not be determined"
            )

    @property
    def segments(self) -> frozenset[int] | None:
        if self.min_voxel_value is None or self.max_voxel_value is None:
            return None
        if (
            self.max_voxel_value - self.min_voxel_value
            >= MAXIMUM_SEGMENTS_LENGTH
        ):
            return None

        return frozenset(range(self.min_voxel_value, self.max_voxel_value + 1))


def _get_tag_value(tags, tag):
    try:
        return tags[tag].value
    except (KeyError, AttributeError):
        return None


def _get_voxel_spacing_mm(tags, tag):
    """
    Calculate the voxel spacing in mm.

    Use the set of tags from the tiff image to calculate the spacing for a
    particular dimension. Supports INCH and CENTIMETER resolution units.

    Parameters
    ----------
    tags
        The collection of tags from the tif file.
    tag
        The tag that contains the resolution of the tif file along the
        dimension of interest.

    Raises
    ------
    ValidationError
        Raised if an unrecognised resolution unit is used.


    Returns
    -------
        The voxel spacing in mm.

    """
    try:
        # Resolution is a tuple of the number of pixels and the length of the
        # image in cm or inches, depending on resolution unit
        resolution_unit = _get_tag_value(tags, "ResolutionUnit").name
        resolution = _get_tag_value(tags, tag)
        if resolution_unit == "INCH":
            return 25.4 / (resolution[0] / resolution[1])
        elif resolution_unit == "CENTIMETER":
            return 10 / (resolution[0] / resolution[1])
        raise ValidationError(
            f"Invalid resolution unit {resolution_unit}" f" in tiff file"
        )
    except (ZeroDivisionError, TypeError, IndexError) as e:
        raise ValidationError("Invalid resolution in tiff file") from e


def _extract_openslide_properties(
    *, gc_file: GrandChallengeTiffFile, image
) -> GrandChallengeTiffFile:
    if not gc_file.voxel_width_mm and "openslide.mpp-x" in image.properties:
        gc_file.voxel_width_mm = (
            float(image.properties["openslide.mpp-x"]) / 1000
        )
    if not gc_file.voxel_height_mm and "openslide.mpp-y" in image.properties:
        gc_file.voxel_height_mm = (
            float(image.properties["openslide.mpp-y"]) / 1000
        )
    if (
        not gc_file.image_height
        and "openslide.level[0].height" in image.properties
    ):
        gc_file.image_height = int(
            image.properties["openslide.level[0].height"]
        )

    if (
        not gc_file.image_width
        and "openslide.level[0].width" in image.properties
    ):
        gc_file.image_width = int(image.properties["openslide.level[0].width"])

    if (
        not gc_file.resolution_levels
        and "openslide.level-count" in image.properties
    ):
        gc_file.resolution_levels = int(
            image.properties["openslide.level-count"]
        )
    return gc_file


def _extract_tags(
    *,
    gc_file: GrandChallengeTiffFile,
    pages: tifffile.tifffile.TiffPages,
    byteorder: str,
) -> GrandChallengeTiffFile:
    """
    Extracts tags form a tiff file loaded with tifffile for use in grand challenge
    :param pages: The pages and tags from tifffile
    :return: The GrandChallengeTiffFile with the properties defined in the tags
    """
    if not pages:
        return gc_file

    tags = pages[0].tags

    gc_file.resolution_levels = len(pages)

    gc_file.color_space = _get_color_space(
        color_space_string=_get_tag_value(
            tags, "PhotometricInterpretation"
        ).name
    )
    gc_file.image_width = _get_tag_value(tags, "ImageWidth")
    gc_file.image_height = _get_tag_value(tags, "ImageLength")

    #  some formats like the Philips tiff don't have the spacing in their tags,
    #  we retrieve them later with OpenSlide
    if "XResolution" in tags:
        gc_file.voxel_width_mm = _get_voxel_spacing_mm(tags, "XResolution")
        gc_file.voxel_height_mm = _get_voxel_spacing_mm(tags, "YResolution")

    get_min_max_sample_value(tags=tags, gc_file=gc_file, byteorder=byteorder)

    return gc_file


def _get_color_space(*, color_space_string) -> ColorSpace | None:
    color_space_string = color_space_string.upper()

    if color_space_string == "MINISBLACK":
        color_space = ColorSpace.GRAY
    else:
        try:
            color_space = ColorSpace[color_space_string]
        except KeyError:
            return None

    return color_space


def get_min_max_sample_value(*, tags, gc_file, byteorder):
    samples_per_pixel = _get_tag_value(tags, "SamplesPerPixel")
    sample_format = _get_tag_value(tags, "SampleFormat")
    if samples_per_pixel != 1 or sample_format not in (None, 1, 2):
        return

    signed = sample_format == 2
    byteorder = {"<": "little", ">": "big"}.get(byteorder, "big")

    def get_voxel_value(first_tag, second_tag):
        voxel_value = _get_tag_value(tags, first_tag) or _get_tag_value(
            tags, second_tag
        )
        if isinstance(voxel_value, bytes):
            return int.from_bytes(
                voxel_value, byteorder=byteorder, signed=signed
            )
        return voxel_value

    gc_file.min_voxel_value = get_voxel_value(
        "MinSampleValue", "SMinSampleValue"
    )
    gc_file.max_voxel_value = get_voxel_value(
        "MaxSampleValue", "SMaxSampleValue"
    )


def _load_with_tiff(
    *, gc_file: GrandChallengeTiffFile
) -> GrandChallengeTiffFile:
    tiff_file = tifffile.TiffFile(str(gc_file.path.absolute()))
    gc_file = _extract_tags(
        gc_file=gc_file, pages=tiff_file.pages, byteorder=tiff_file.byteorder
    )
    return gc_file


def _load_with_openslide(
    *, gc_file: GrandChallengeTiffFile
) -> GrandChallengeTiffFile:
    import openslide

    open_slide_file = openslide.open_slide(str(gc_file.path.absolute()))
    gc_file = _extract_openslide_properties(
        gc_file=gc_file, image=open_slide_file
    )
    return gc_file


mirax_pattern = r"INDEXFILE\s?=\s?.|FILE_\d+\s?=\s?."
# vms (and vmu) files contain key value pairs, where the ImageFile keys
# can have the following format:
# ImageFile =, ImageFile(x,y) ImageFile(z,x,y)
vms_pattern = (
    r"ImageFile(\(\d*,\d*(,\d)?\))?\s?=\s?."
    r"|MapFile\s?=\s?."
    r"|OptimisationFile\s?=\s?."
    r"|MacroImage\s?=\s?."
)


def _get_mrxs_files(mrxs_file: Path):
    # Find Slidedat.ini, which provides us with all the other file names
    name, _ = os.path.splitext(mrxs_file.name)
    slide_dat = mrxs_file.parent / name / "Slidedat.ini"
    file_matched = []
    if not slide_dat.exists():
        slide_dat = [
            f
            for f in slide_dat.parent.iterdir()
            if f.name.lower() == slide_dat.name.lower()
        ][0]
    with open(slide_dat) as f:
        lines = [
            line for line in f.readlines() if re.match(mirax_pattern, line)
        ]
        for line in lines:
            original_name = line.split("=")[1].strip()
            if os.path.exists(slide_dat.parent / original_name):
                file_matched.append(slide_dat.parent / original_name)
        file_matched.append(slide_dat)
    return file_matched


def _get_vms_files(vms_file: Path):
    file_matched = []
    with open(str(vms_file.absolute())) as f:
        lines = [line for line in f.readlines() if re.match(vms_pattern, line)]
        for line in lines:
            original_name = line.split("=")[1].strip()
            if os.path.exists(vms_file.parent / original_name):
                file_matched.append(vms_file.parent / original_name)
        return file_matched


def _convert(
    *,
    files: list[Path],
    associated_files_getter: Callable[[Path], list[Path]] | None,
    output_directory: Path,
    file_errors: dict[Path, list[str]],
) -> list[GrandChallengeTiffFile]:
    converted_files: list[GrandChallengeTiffFile] = []
    associated_files: list[Path] = []

    for file in files:
        try:
            gc_file = GrandChallengeTiffFile(file)

            if associated_files_getter:
                associated_files = associated_files_getter(gc_file.path)

            tiff_file = _convert_to_tiff(
                path=file,
                pk=gc_file.pk,
                output_directory=output_directory,
            )
        except Exception:
            file_errors[file].append(
                format_error(f"Could not convert file to TIFF: {file.name}")
            )
            continue
        else:
            gc_file.path = tiff_file
            gc_file.associated_files = associated_files
            gc_file.associated_files.append(file)
            converted_files.append(gc_file)

    return converted_files


def _convert_to_tiff(*, path: Path, pk: UUID, output_directory: Path) -> Path:
    import pyvips

    major = pyvips.base.version(0)
    minor = pyvips.base.version(1)

    if not (major > 8 or (major == 8 and minor >= 10)):
        raise RuntimeError(
            f"libvips {major}.{minor} is too old - requires >= 8.10"
        )

    new_file_name = output_directory / path.name / f"{pk}.tif"
    new_file_name.parent.mkdir()

    image = pyvips.Image.new_from_file(
        str(path.absolute()), access="sequential"
    )

    image.write_to_file(
        str(new_file_name.absolute()),
        tile=True,
        pyramid=True,
        bigtiff=True,
        compression="jpeg",
        Q=70,
    )

    return new_file_name


def _convert_dicom_wsi_dir(
    gc_file: GrandChallengeTiffFile,
    file: Path,
    output_directory: Path,
    file_errors: dict[Path, list[str]],
):
    wsidicom_dir = file.parent
    try:
        new_file_name = output_directory / file.name / f"{gc_file.pk}.tif"
        new_file_name.parent.mkdir()

        wsi_dcm_to_tiff(wsidicom_dir, new_file_name)
    except Exception:
        file_errors[file].append(
            format_error(f"Could not convert dicom-wsi to TIFF: {file.name}")
        )
    else:
        gc_file.path = new_file_name
        gc_file.associated_files = [
            f for f in wsidicom_dir.iterdir() if f.is_file()
        ]
    return gc_file


def _convert_isyntax_file(
    gc_file: GrandChallengeTiffFile,
    file: Path,
    output_directory: Path,
    file_errors: dict[Path, list[str]],
):
    try:
        new_file_name = output_directory / file.name / f"{gc_file.pk}.tif"
        new_file_name.parent.mkdir()

        wsi_isyntax_to_tiff(file, new_file_name)
    except Exception:
        file_errors[file].append(
            format_error(f"Could not convert iSyntax to TIFF: {file.name}")
        )
    else:
        gc_file.path = new_file_name
        gc_file.associated_files = [file]
    return gc_file


def _find_valid_dicom_wsi_files(
    files: set[Path], file_errors: DefaultDict[Path, list[str]]
):
    """
    Gets the headers for all dicom files on path and validates them.

    Parameters
    ----------
    files
        Paths images that were uploaded during an upload session.

    file_errors
        Dictionary in which reading errors are recorded per file

    Returns
    -------
    A dictionary with filename as key, and all other files belonging to that study
    as value

    """
    # Try and get dicom files; ignore errors
    dicom_errors = file_errors.copy()
    studies = get_dicom_headers_by_study(files=files, file_errors=dicom_errors)
    result: dict[Path, list[Path]] = {}

    for key in studies:
        headers = studies[key]["headers"]
        if not headers:
            continue

        if not all(
            header["data"].SOPClassUID == DICOM_WSI_STORAGE_ID
            for header in headers
        ):
            for d in headers:
                file_errors[d["file"]].append(
                    format_error("Non-WSI-DICOM not supported by TIFF builder")
                )
        else:
            result[Path(headers[0]["file"])] = [
                Path(h["file"]) for h in headers[1:]
            ]

    def associated_files(file_path: Path):
        return result[file_path]

    return list(result.keys()), associated_files


def _find_valid_isyntax_wsi_files(
    files: set[Path], file_errors: DefaultDict[Path, list[str]]
):
    """
    Gets the headers for all isyntax files on path and validates them.

    Parameters
    ----------
    files
        Paths images that were uploaded during an upload session.

    file_errors
        Dictionary in which reading errors are recorded per file

    Returns
    -------
    A dictionary with filename as key, and all other files belonging to that study
    as value

    """
    result: dict[Path, list[Path]] = {}

    isyntax_files = [
        file for file in files if file.suffix.casefold() == ".isyntax"
    ]
    for isyntax_file in isyntax_files:
        try:
            if not _has_isyntax:
                raise ImportError("Install pyisyntax to convert isyntax files")
            with ISyntax.open(isyntax_file) as image:
                wsi = image.wsi
                if not wsi.level_count:
                    file_errors[isyntax_file].append(
                        format_error("No levels found in iSyntax file")
                    )
                else:
                    result[isyntax_file] = [isyntax_file]
        except Exception:
            file_errors[isyntax_file].append(
                format_error("Could not open iSyntax file")
            )

    def associated_files(file_path: Path):
        return result[file_path]

    return list(result.keys()), associated_files


def _load_gc_files(
    *,
    files: set[Path],
    output_directory: Path,
    file_errors: DefaultDict[Path, list[str]],
) -> list[GrandChallengeTiffFile]:
    loaded_files: list[GrandChallengeTiffFile] = []

    complex_file_handlers = {
        ".mrxs": _get_mrxs_files,
        ".vms": _get_vms_files,
        ".vmu": _get_vms_files,
        ".svs": None,
        ".ndpi": None,
        ".scn": None,
        ".bif": None,
    }

    dicom_files, handler = _find_valid_dicom_wsi_files(files, file_errors)
    for dicom_file in dicom_files:
        gc_file = GrandChallengeTiffFile(dicom_file)
        gc_file = _convert_dicom_wsi_dir(
            gc_file=gc_file,
            file=dicom_file,
            output_directory=output_directory,
            file_errors=file_errors,
        )
        loaded_files.append(gc_file)

    isyntax_files, handler = _find_valid_isyntax_wsi_files(files, file_errors)
    for isyntax_file in isyntax_files:
        gc_file = GrandChallengeTiffFile(isyntax_file)
        gc_file = _convert_isyntax_file(
            gc_file=gc_file,
            file=isyntax_file,
            output_directory=output_directory,
            file_errors=file_errors,
        )
        loaded_files.append(gc_file)

    for ext, handler in complex_file_handlers.items():
        complex_files = [file for file in files if file.suffix.lower() == ext]
        if len(complex_files) > 0:
            converted_files = _convert(
                files=complex_files,
                associated_files_getter=handler,
                output_directory=output_directory,
                file_errors=file_errors,
            )
            loaded_files += converted_files

    # don't handle files that are associated files
    for file in files:
        if not any(g.path == file for g in loaded_files) and not any(
            file in g.associated_files
            for g in loaded_files
            if g.associated_files is not None
        ):
            gc_file = GrandChallengeTiffFile(file)
            loaded_files.append(gc_file)

    return loaded_files


def image_builder_tiff(  # noqa: C901
    *, files: set[Path]
) -> Iterator[TIFFImage]:
    file_errors: DefaultDict[Path, list[str]] = defaultdict(list)

    with TemporaryDirectory() as output_directory:
        loaded_files = _load_gc_files(
            files=files,
            output_directory=Path(output_directory),
            file_errors=file_errors,
        )

        for gc_file in loaded_files:
            # try and load image with tiff file
            try:
                gc_file = _load_with_tiff(gc_file=gc_file)
            except Exception:
                file_errors[gc_file.path].append(
                    format_error("Could not open file with tifffile.")
                )

            # try and load image with open slide
            try:
                gc_file = _load_with_openslide(gc_file=gc_file)
            except Exception:
                file_errors[gc_file.path].append(
                    format_error("Could not open file with OpenSlide.")
                )

            # validate
            try:
                gc_file.validate()
                if gc_file.color_space is None:
                    # TODO This needs to be solved by refactoring of
                    # GrandChallengeTiffFile
                    raise RuntimeError("Color space not found")
            except ValidationError as e:
                file_errors[gc_file.path].append(
                    format_error(f"Validation error: {e}.")
                )
                continue

            if gc_file.associated_files:
                consumed_files = {
                    f.absolute() for f in gc_file.associated_files
                }
            else:
                consumed_files = {gc_file.path.absolute()}

            yield TIFFImage(
                file=gc_file.path,
                name=gc_file.path.name,
                consumed_files=consumed_files,
                width=gc_file.image_width,
                height=gc_file.image_height,
                voxel_width_mm=gc_file.voxel_width_mm,
                voxel_height_mm=gc_file.voxel_height_mm,
                resolution_levels=gc_file.resolution_levels,
                color_space=gc_file.color_space,
                segments=gc_file.segments,
            )

    if file_errors:
        raise UnconsumedFilesException(file_errors=file_errors)



class OCTVolumeWithMetaData:
    """Class to hold the OCT volume and any related metadata, and enable viewing and saving.

    Attributes:
        volume (list of np.array): All the volume's b-scans.
        laterality (str): Left or right eye.
        patient_id (str): Patient ID.
        DOB (str): Patient date of birth.
        num_slices: Number of b-scans present in volume.
    """

    def __init__(
        self, volume, laterality=None, patient_id=None, patient_dob=None
    ):
        self.volume = volume
        self.laterality = laterality
        self.patient_id = patient_id
        self.DOB = patient_dob
        self.num_slices = len(self.volume)



import shutil
import zlib
from pathlib import Path

import pytest
import SimpleITK

from panimg.exceptions import ValidationError
from panimg.image_builders.metaio_utils import (
    ADDITIONAL_HEADERS,
    FLOAT_OR_FLOAT_ARRAY_MATCH_REGEX,
    load_sitk_image,
    parse_mh_header,
)
from tests import RESOURCE_PATH

MHD_WINDOW_DIR = RESOURCE_PATH / "mhd_window"


def test_parse_header_valid_4d_mhd():
    headers = parse_mh_header(RESOURCE_PATH / "image10x11x12x13.mhd")
    assert headers == {
        "ObjectType": "Image",
        "NDims": "4",
        "BinaryData": "True",
        "BinaryDataByteOrderMSB": "False",
        "CompressedData": "True",
        "CompressedDataSize": "40",
        "TransformMatrix": "1 0 0 0 0 1 0 0 0 0 1 0 0 0 0 1",
        "Offset": "-131 -99 -917 0",
        "CenterOfRotation": "0 0 0 0",
        "AnatomicalOrientation": "????",
        "ElementSpacing": "0.42899999999999999 0.42899999999999999 0.5 1",
        "DimSize": "10 11 12 13",
        "ElementType": "MET_UCHAR",
        "ElementDataFile": "image10x11x12x13.zraw",
    }


def test_parse_header_valid_mhd():
    headers = parse_mh_header(RESOURCE_PATH / "image10x10x10.mhd")
    assert headers == {
        "ObjectType": "Image",
        "NDims": "3",
        "BinaryData": "True",
        "BinaryDataByteOrderMSB": "False",
        "CompressedData": "True",
        "CompressedDataSize": "7551",
        "TransformMatrix": "1 0 0 0 1 0 0 0 1",
        "Offset": "0 0 0",
        "CenterOfRotation": "0 0 0",
        "AnatomicalOrientation": "RAI",
        "ElementSpacing": "1 1 1",
        "DimSize": "10 10 10",
        "ElementType": "MET_DOUBLE",
        "ElementDataFile": "image10x10x10.zraw",
    }


def test_parse_header_valid_mhd_with_extra_fields():
    headers = parse_mh_header(RESOURCE_PATH / "image10x10x10-extra-stuff.mhd")
    assert headers == {
        "ObjectType": "Image",
        "NDims": "3",
        "BinaryData": "True",
        "BinaryDataByteOrderMSB": "False",
        "CompressedData": "True",
        "CompressedDataSize": "7551",
        "TransformMatrix": "1 0 0 0 1 0 0 0 1",
        "Offset": "0 0 0",
        "CenterOfRotation": "0 0 0",
        "AnatomicalOrientation": "RAI",
        "ElementSpacing": "1 1 1",
        "DimSize": "10 10 10",
        "ElementType": "MET_DOUBLE",
        "ElementDataFile": "image10x10x10.zraw",
        "Some_values": '"Huh? \u2713\U0001f604"',
    }


def assert_sitk_img_equivalence(
    img: SimpleITK.Image, img_ref: SimpleITK.Image
):
    assert img.GetDimension() == img_ref.GetDimension()
    assert img.GetDirection() == img.GetDirection()
    assert img.GetSize() == img_ref.GetSize()
    assert img.GetOrigin() == img_ref.GetOrigin()
    assert img.GetSpacing() == img_ref.GetSpacing()
    assert (
        img.GetNumberOfComponentsPerPixel()
        == img_ref.GetNumberOfComponentsPerPixel()
    )
    assert img.GetPixelIDValue() == img_ref.GetPixelIDValue()
    assert img.GetPixelIDTypeAsString() == img_ref.GetPixelIDTypeAsString()


@pytest.mark.parametrize(
    "image",
    (
        RESOURCE_PATH / "image10x11x12x13.mhd",
        RESOURCE_PATH / "image10x11x12x13.mha",
    ),
)
def test_writing_4d_mhd_produces_same_results(tmpdir, image: Path):
    def assert_img_properties(img: SimpleITK.Image):
        assert img.GetDimension() == 4
        assert img.GetWidth() == 10
        assert img.GetHeight() == 11
        assert img.GetDepth() == 12
        assert img.GetSize()[-1] == 13

    img_ref = load_sitk_image(image)
    assert_img_properties(img_ref)
    copypath = Path(tmpdir / "temp4d.mhd")
    SimpleITK.WriteImage(img_ref, str(copypath), True)
    img = load_sitk_image(copypath)
    assert_img_properties(img)
    assert_sitk_img_equivalence(img, img_ref)


def test_4d_mh_loader_without_datafile_fails(tmpdir):
    src = RESOURCE_PATH / "image10x11x12x13.mhd"
    dest = Path(tmpdir) / src.name
    shutil.copy(str(src), str(dest))
    with pytest.raises(RuntimeError, match="File cannot be read"):
        load_sitk_image(dest)


def test_4d_mh_loader_with_invalid_data_type_fails(tmpdir):
    sources = [
        RESOURCE_PATH / "image10x11x12x13.mhd",
        RESOURCE_PATH / "image10x11x12x13.zraw",
    ]
    targets = []
    for src in sources:
        dest = Path(tmpdir) / src.name
        targets.append(dest)
        shutil.copy(str(src), str(dest))
    tmp_header_file = targets[0]
    with open(str(tmp_header_file)) as f:
        modified_header = f.read().replace("MET_UCHAR", "MET_OTHER")
    with open(str(tmp_header_file), "w") as f:
        f.write(modified_header)
    with pytest.raises(RuntimeError, match="Unknown PixelType"):
        load_sitk_image(tmp_header_file)


def test_4d_mh_loader_with_uncompressed_data(tmpdir):
    sources = [
        RESOURCE_PATH / "image10x11x12x13.mhd",
        RESOURCE_PATH / "image10x11x12x13.zraw",
    ]
    targets = []
    for src in sources:
        dest = Path(tmpdir) / src.name
        targets.append(dest)
        shutil.copy(str(src), str(dest))
    tmp_header_file, tmp_data_file = targets
    with open(str(tmp_header_file)) as f:
        modified_header = f.read().replace(
            "CompressedData = True", "CompressedData = False"
        )
    with open(str(tmp_header_file), "w") as f:
        f.write(modified_header)
    with open(str(tmp_data_file), "rb") as f:
        data = zlib.decompress(f.read())
    with open(str(tmp_data_file), "wb") as f:
        f.write(data)
    load_sitk_image(tmp_header_file)


def test_4d_mh_loader_with_more_than_4_dimensions_fails():
    with pytest.raises(NotImplementedError):
        load_sitk_image(RESOURCE_PATH / "image4x4x4x4x4.mha")


@pytest.mark.parametrize(
    "test_img",
    ["image10x11x12x13-extra-stuff.mhd", "image3x4-extra-stuff.mhd"],
)
def test_load_sitk_image_with_additional_meta_data(tmpdir, test_img: str):
    src = RESOURCE_PATH / test_img
    sitk_image = load_sitk_image(src)
    for key in sitk_image.GetMetaDataKeys():
        assert key in ADDITIONAL_HEADERS
        assert ADDITIONAL_HEADERS[key].match(sitk_image.GetMetaData(key))
    assert "Bogus" not in sitk_image.GetMetaDataKeys()


@pytest.mark.parametrize(
    "test_img, center, width",
    [
        ("image3x4.mhd", "20.5", "200.5"),
        ("image3x4-len-1.mhd", "[20.5]", "[200.5]"),
        ("image3x4-len-2.mhd", "[10.5, 20.5]", "[100.5, 200.5]"),
    ],
)
def test_load_sitk_image_with_various_window_formats(
    test_img: str,
    center: float | list[float],
    width: float | list[float],
):
    src = MHD_WINDOW_DIR / test_img
    sitk_image = load_sitk_image(src)

    assert center == sitk_image.GetMetaData("WindowCenter")
    assert width == sitk_image.GetMetaData("WindowWidth")


@pytest.mark.parametrize(
    "test_img, error_msg",
    [
        ("image3x4-center-gt-width.mhd", "equal number"),
        ("image3x4-width-gt-center.mhd", "equal number"),
        ("image3x4-only-width-array.mhd", "different format"),
        ("image3x4-only-center-array.mhd", "different format"),
        ("image3x4-len-0.mhd", "does not match pattern"),
    ],
)
def test_load_sitk_image_with_faulty_window_formats(
    test_img: str, error_msg: str
):
    src = MHD_WINDOW_DIR / "errors" / test_img
    with pytest.raises(ValidationError, match=error_msg):
        load_sitk_image(src)


@pytest.mark.parametrize("test_img", ["image10x11x12x13.mhd", "image3x4.mhd"])
@pytest.mark.parametrize(
    ["key", "value"],
    [
        ("Exposures", "1 2 3e-5 4 5.2 6 7 8 9 10 11 12 13.0 e"),
        ("Exposures", "1 2 3e-5 4 5.2 6 7 8 9 10 11 12"),
        ("t0", "1 2"),
        ("t1", "string"),
        (
            "ContentTimes",
            ("245959.999 000000.000 111111.111 " * 4) + "121212.333",
        ),
        ("ContentTimes", "1 2 3e-5 4 5.2 6 7 8 9 10 11 12 13.0 14.0"),
    ],
)
def test_load_sitk_image_with_corrupt_additional_meta_data_fails(
    tmpdir, test_img: str, key: str, value: str
):
    src = RESOURCE_PATH / "image10x11x12x13.mhd"
    dest = Path(tmpdir) / src.name
    shutil.copy(str(src), str(dest))
    with open(str(dest)) as f:
        lines = f.readlines()
    lines.insert(-1, f"{key} = {value}\n")
    with open(str(dest), "w") as f:
        f.writelines(lines)
    with pytest.raises(ValidationError):
        load_sitk_image(dest)


def test_fail_on_invalid_utf8():
    with pytest.raises(ValidationError):
        parse_mh_header(RESOURCE_PATH / "invalid_utf8.mhd")


def test_too_many_headers_file(tmpdir):
    # Generate test file...
    test_file_path = Path(tmpdir) / "test.mhd"
    with open(test_file_path, "w", encoding="utf-8") as f:
        for i in range(1000000):
            f.write(f"key{i} = {i}\n")

    with pytest.raises(ValidationError):
        parse_mh_header(test_file_path)


def test_line_too_long(tmpdir):
    # Generate test file...
    test_file_path = Path(tmpdir) / "test.mhd"
    with open(test_file_path, "w", encoding="utf-8") as f:
        f.write("key = ")
        for i in range(1000000):
            f.write(f"{i}")

    with pytest.raises(ValidationError):
        parse_mh_header(test_file_path)


def test_does_not_choke_on_empty_file(tmpdir):
    # Generate test file...
    test_file_path = Path(tmpdir) / "test.mhd"
    with open(test_file_path, "w", encoding="utf-8") as f:
        f.write("\n")

    assert parse_mh_header(test_file_path) == {}


@pytest.mark.parametrize(
    "string",
    [
        "1",
        "1.001",
        "-1e10",
        "+1e10",
        "00400",
        "[1]",
        "[1.1]",
        "[1, 1]",
        "[1.1, 0]",
        "[0, 1.1]",
        "[-1e10, +100, 00400]",
        "[1, 2, 3, 4]",
        "[1,2,3,4]",
    ],
)
def test_regex_matching_float_arrays(string: str):
    assert FLOAT_OR_FLOAT_ARRAY_MATCH_REGEX.match(string)


@pytest.mark.parametrize(
    "string",
    [
        "",
        ".",
        "[]",
        "1,",
        "1,  ",
        "[1  ]",
        "[  1]",
        "[1",
        "1]",
        "[1,]",
        "[ 1]",
        "[ , ]",
        "[,]",
        "[1, 2, 3, 4,]",
    ],
)
def test_regex_non_matching_faulty_array_and_floats(string: str):
    assert not FLOAT_OR_FLOAT_ARRAY_MATCH_REGEX.match(string)



from pathlib import Path

import pytest
import SimpleITK
from pytest import approx

from panimg.image_builders.metaio_utils import load_sitk_image
from panimg.models import ColorSpace, SimpleITKImage
from tests import RESOURCE_PATH


@pytest.mark.parametrize(
    "image",
    (
        RESOURCE_PATH / "image3x4.mhd",
        RESOURCE_PATH / "image3x4-extra-stuff.mhd",
        RESOURCE_PATH / "image5x6x7.mhd",
        RESOURCE_PATH / "image10x10x10.mhd",
        RESOURCE_PATH / "image10x10x10.mha",
        RESOURCE_PATH / "image10x10x10-extra-stuff.mhd",
        RESOURCE_PATH / "image10x11x12x13.mhd",
        RESOURCE_PATH / "image10x11x12x13.mhd",
        RESOURCE_PATH / "image10x11x12x13-extra-stuff.mhd",
        RESOURCE_PATH / "image128x256RGB.mhd",
        RESOURCE_PATH / "image128x256x3RGB.mhd",
        RESOURCE_PATH / "image128x256x4RGB.mhd",
    ),
)
def test_convert_itk_to_internal(image: Path):
    def assert_img_properties(
        img: SimpleITK.Image, internal_image: SimpleITKImage
    ):
        color_space = {
            1: ColorSpace.GRAY,
            3: ColorSpace.RGB,
            4: ColorSpace.RGBA,
        }

        assert internal_image.color_space == color_space.get(
            img.GetNumberOfComponentsPerPixel()
        )
        if img.GetDimension() == 4:
            assert internal_image.timepoints == img.GetSize()[-1]
        else:
            assert internal_image.timepoints is None
        if img.GetDepth():
            assert internal_image.depth == img.GetDepth()
            assert internal_image.voxel_depth_mm == img.GetSpacing()[2]
        else:
            assert internal_image.depth is None
            assert internal_image.voxel_depth_mm is None

        assert internal_image.width == img.GetWidth()
        assert internal_image.height == img.GetHeight()
        assert internal_image.voxel_width_mm == approx(img.GetSpacing()[0])
        assert internal_image.voxel_height_mm == approx(img.GetSpacing()[1])

    img_ref = load_sitk_image(image)
    internal_image = SimpleITKImage(
        name=image.name,
        image=img_ref,
        consumed_files=set(),
        spacing_valid=True,
    )
    assert_img_properties(img_ref, internal_image)



from pathlib import Path
from tempfile import TemporaryDirectory

import pytest
from tifffile import TiffFile

from panimg.contrib.wsi_isyntax_to_tiff.isyntax_to_tiff import isyntax_to_tiff


def test_isyntax_to_tiff(downloaded_isyntax_image):
    if not downloaded_isyntax_image.exists():
        pytest.xfail(
            reason="iSyntax resource is not available. "
            "To run this test, download the file using --download-files."
        )

    from isyntax import ISyntax

    with TemporaryDirectory() as output_directory:
        converted_tiff = Path(output_directory) / "output.tiff"
        isyntax_to_tiff(downloaded_isyntax_image, converted_tiff)
        with ISyntax.open(downloaded_isyntax_image) as o_tif:
            with TiffFile(converted_tiff) as c_tif:
                assert o_tif.level_count == len(c_tif.pages)
                for i in range(0, o_tif.wsi.level_count):
                    assert (
                        o_tif.wsi.get_level(i).height
                        == c_tif.pages[i].shape[0]
                    )
                    assert (
                        o_tif.wsi.get_level(i).width == c_tif.pages[i].shape[1]
                    )
                    assert o_tif.tile_height == c_tif.pages[i].tile[0]
                    assert o_tif.tile_width == c_tif.pages[i].tile[1]



import logging
import sys
from collections.abc import Iterable
from importlib.metadata import version
from pathlib import Path
from uuid import UUID

import click
from pydantic import RootModel

from panimg import convert, post_process
from panimg.image_builders import (
    DEFAULT_IMAGE_BUILDERS,
    IMAGE_BUILDER_OPTIONS_TO_IMPLEMENTATION,
    ImageBuilderOptions,
)
from panimg.models import (
    ImageType,
    PanImgFile,
    PanImgResult,
    PostProcessorResult,
)
from panimg.post_processors import (
    DEFAULT_POST_PROCESSORS,
    POST_PROCESSOR_OPTIONS_TO_IMPLEMENTATION,
    PostProcessorOptions,
)
from panimg.types import ImageBuilder

logger = logging.getLogger(__name__)

panimg_version = version("panimg")


@click.group(context_settings=dict(help_option_names=["-h", "--help"]))
@click.version_option(panimg_version, "-v", "--version")
def cli():
    pass


@cli.command(name="convert", short_help="Convert a directory of image files")
@click.option("-v", "--verbose", count=True)
@click.option(
    "--input-dir",
    type=click.Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
        readable=True,
        writable=False,
        resolve_path=True,
        path_type=Path,
    ),
    required=True,
)
@click.option(
    "--output-dir",
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        readable=True,
        writable=True,
        resolve_path=True,
        path_type=Path,
    ),
    required=True,
)
@click.option(
    "--post-processor", type=click.Choice(PostProcessorOptions), multiple=True
)
@click.option("--no-post-processing", is_flag=True, default=False)
@click.option(
    "--image-builder", type=click.Choice(ImageBuilderOptions), multiple=True
)
def convert_cli(
    *,
    verbose: int,
    input_dir: Path,
    output_dir: Path,
    post_processor: tuple[PostProcessorOptions],
    no_post_processing: bool,
    image_builder: tuple[ImageBuilderOptions],
):
    _setup_verbosity(level=verbose)

    if output_dir.exists():
        raise click.exceptions.UsageError(
            f"Output directory {output_dir} already exists"
        )
    else:
        output_dir.mkdir(parents=True)

    if post_processor and no_post_processing:
        raise click.ClickException(
            "--no-post-processing and --post-processor cannot be used together"
        )
    elif no_post_processing:
        processors = []
    elif post_processor:
        processors = [
            POST_PROCESSOR_OPTIONS_TO_IMPLEMENTATION[processor]
            for processor in post_processor
        ]
    else:
        processors = DEFAULT_POST_PROCESSORS

    if image_builder:
        builders: Iterable[ImageBuilder] = [
            IMAGE_BUILDER_OPTIONS_TO_IMPLEMENTATION[builder]
            for builder in image_builder
        ]
    else:
        builders = DEFAULT_IMAGE_BUILDERS

    logger.info(
        f"Converting {input_dir} to {output_dir} using {builders} and {processors}"
    )

    result = convert(
        input_directory=input_dir,
        output_directory=output_dir,
        builders=builders,
        post_processors=processors,
    )

    click.echo(RootModel[PanImgResult](result).model_dump_json())


@cli.command(name="post-process", short_help="Post process an image file")
@click.option("-v", "--verbose", count=True)
@click.option("--image-id", type=UUID, required=True)
@click.option("--image-type", type=click.Choice(ImageType), required=True)
@click.option(
    "--input-file",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        writable=False,
        resolve_path=True,
        path_type=Path,
    ),
    required=True,
)
@click.option(
    "--post-processor", type=click.Choice(PostProcessorOptions), multiple=True
)
def post_process_cli(
    *,
    verbose: int,
    image_id: UUID,
    image_type: ImageType,
    input_file: Path,
    post_processor: tuple[PostProcessorOptions],
):
    _setup_verbosity(level=verbose)

    if post_processor:
        processors = [
            POST_PROCESSOR_OPTIONS_TO_IMPLEMENTATION[processor]
            for processor in post_processor
        ]
    else:
        processors = DEFAULT_POST_PROCESSORS

    logger.info(
        f"Post processing {input_file} ({image_type}) using {processors}"
    )

    result = post_process(
        image_files={
            PanImgFile(
                image_id=image_id, image_type=image_type, file=input_file
            )
        },
        post_processors=processors,
    )

    click.echo(RootModel[PostProcessorResult](result).model_dump_json())


class MaxLevelFilter(logging.Filter):
    def __init__(self, max_level):
        super().__init__()
        self.max_level = max_level

    def filter(self, record):
        return record.levelno < self.max_level


def _setup_verbosity(*, level: int):
    if level == 0:
        logger.setLevel(logging.WARNING)
    elif level == 1:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.DEBUG)

    # stdout handler: DEBUG and INFO (levels < WARNING)
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(logging.NOTSET)
    stdout_handler.addFilter(MaxLevelFilter(logging.WARNING))

    # stderr handler: WARNING and above
    stderr_handler = logging.StreamHandler(sys.stderr)
    stderr_handler.setLevel(logging.WARNING)

    logger.addHandler(stdout_handler)
    logger.addHandler(stderr_handler)



import os
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from panimg.image_builders.fallback import format_error, image_builder_fallback
from panimg.models import ColorSpace
from panimg.panimg import _build_files
from tests import RESOURCE_PATH


@pytest.mark.parametrize(
    "src,colorspace",
    (
        (RESOURCE_PATH / "test_grayscale.jpg", ColorSpace.GRAY),
        (RESOURCE_PATH / "test_rgb.jpg", ColorSpace.RGB),
        (RESOURCE_PATH / "test_grayscale.png", ColorSpace.GRAY),
        (RESOURCE_PATH / "test_rgb.png", ColorSpace.RGB),
        (RESOURCE_PATH / "test_rgba.png", ColorSpace.RGBA),
    ),
)
def test_image_builder_fallback(tmpdir, src, colorspace):
    dest = Path(tmpdir) / src.name
    shutil.copy(str(src), str(dest))
    files = {Path(d[0]).joinpath(f) for d in os.walk(tmpdir) for f in d[2]}
    with TemporaryDirectory() as output:
        result = _build_files(
            builder=image_builder_fallback,
            files=files,
            output_directory=output,
        )

    assert result.consumed_files == {dest}
    assert len(result.new_images) == 1
    image = result.new_images.pop()
    assert image.color_space == colorspace
    assert image.voxel_width_mm is None
    assert image.voxel_height_mm is None
    assert image.voxel_depth_mm is None


def test_image_builder_fallback_corrupt_file(tmpdir):
    src = RESOURCE_PATH / "corrupt.png"
    dest = Path(tmpdir) / src.name
    shutil.copy(str(src), str(dest))

    files = {Path(d[0]).joinpath(f) for d in os.walk(tmpdir) for f in d[2]}
    with TemporaryDirectory() as output:
        result = _build_files(
            builder=image_builder_fallback,
            files=files,
            output_directory=output,
        )

    assert result.file_errors == {
        dest: [format_error("Not a valid image file")]
    }
    assert result.consumed_files == set()



from collections.abc import Iterator
from pathlib import Path
from typing import Protocol

from panimg.models import (
    PanImgFile,
    PostProcessorResult,
    SimpleITKImage,
    TIFFImage,
)


class ImageBuilder(Protocol):
    def __call__(  # noqa: E704
        self, *, files: set[Path]
    ) -> Iterator[SimpleITKImage] | Iterator[TIFFImage]: ...


class PostProcessor(Protocol):
    def __call__(  # noqa: E704
        self, *, image_files: set[PanImgFile]
    ) -> PostProcessorResult: ...



import pytest

from panimg.exceptions import UnconsumedFilesException
from panimg.image_builders import DEFAULT_IMAGE_BUILDERS, image_builder_oct
from panimg.panimg import _build_files
from tests import RESOURCE_PATH

EXPECTED_ERROR_MESSAGE = {
    "image_builder_dicom": ["Dicom image builder: could not parse headers"],
    "image_builder_fallback": [
        "Fallback image builder: Not a valid image file"
    ],
    "image_builder_mhd": ["Mhd image builder: Not an ITK file"],
    "image_builder_nifti": ["NifTI image builder: Not a NifTI image file"],
    "image_builder_nrrd": ["NRRD image builder: Not a NRRD image file"],
    "image_builder_oct": [
        "OCT image builder: Not a valid OCT file "
        "(supported formats: .fds,.fda,.e2e)"
    ],
    "image_builder_tiff": [
        "TIFF image builder: Could not open file with tifffile.",
        "TIFF image builder: Could not open file with OpenSlide.",
        "TIFF image builder: Validation error: Not a valid tif: "
        "Image width could not be determined.",
    ],
}


@pytest.mark.parametrize("builder", DEFAULT_IMAGE_BUILDERS)
def test_image_builder_raises_unconsumed_file_exception(tmp_path, builder):
    f = tmp_path / "image.jpg"
    f.write_bytes(b"")

    with pytest.raises(UnconsumedFilesException) as e:
        _ = [*builder(files={f})]

    assert {**e.value.file_errors} == {
        f: EXPECTED_ERROR_MESSAGE[builder.__name__]
    }


@pytest.mark.parametrize("builder", DEFAULT_IMAGE_BUILDERS)
def test_number_of_images_consumed_by_each_builder(tmp_path, builder):
    # Except for the oct builder, each of the builders should only consume one
    # image. The other files should error out.
    files = {
        *(RESOURCE_PATH / "dicom_4d").glob("*.dcm"),
        RESOURCE_PATH / "test_rgb.png",
        RESOURCE_PATH / "image10x10x10.mha",
        RESOURCE_PATH / "nifti" / "image10x11x12.nii.gz",
        RESOURCE_PATH / "nrrd" / "image10x11x12.nrrd",
        RESOURCE_PATH / "valid_tiff.tif",
        RESOURCE_PATH / "oct/fda_minimized.fda",
        RESOURCE_PATH / "oct/fds_minimized.fds",
    }

    result = _build_files(
        builder=builder, files=files, output_directory=tmp_path
    )

    if image_builder_oct == builder:
        assert len(result.new_images) == 2
    else:
        assert len(result.new_images) == 1
    assert len(result.consumed_files) == len(files) - len(result.file_errors)
    assert result.consumed_files | {*result.file_errors} == files



import shutil

import pytest

from panimg import convert
from panimg.image_builders.metaio_utils import (
    ADDITIONAL_HEADERS,
    EXPECTED_HEADERS,
    HEADERS_MATCHING_NUM_TIMEPOINTS,
    HEADERS_WITH_LISTING,
    extract_header_listing,
    parse_mh_header,
)
from panimg.models import ColorSpace
from tests import RESOURCE_PATH


def _convert_images(*, images, tmpdir_factory):
    input_directory = tmpdir_factory.mktemp("input")
    for image in images:
        shutil.copy(RESOURCE_PATH / image, input_directory)

    output_directory = tmpdir_factory.mktemp("output")

    return convert(
        input_directory=input_directory, output_directory=output_directory
    )


def test_image_file_creation(tmpdir_factory):
    images = [
        "image10x10x10.zraw",
        "image10x10x10.mhd",
        "image10x10x10.mha",
        "image10x11x12x13.mhd",
        "image10x11x12x13.zraw",
        "image10x10x10-extra-stuff.mhd",
        "invalid_utf8.mhd",
        "no_image",
        "valid_tiff.tif",
        "invalid_resolutions_tiff.tif",
    ]

    invalid_images = (
        "no_image",
        "invalid_utf8.mhd",
        "invalid_resolutions_tiff.tif",
    )

    result = _convert_images(images=images, tmpdir_factory=tmpdir_factory)

    assert len(result.consumed_files) == len(images) - len(invalid_images)
    assert len(result.new_images) == 5


@pytest.mark.parametrize(
    "images",
    (
        ["image10x11x12x13.mha"],
        ["image10x11x12x13.mhd", "image10x11x12x13.zraw"],
    ),
)
def test_staged_4d_mha_and_4d_mhd_upload(images, tmpdir_factory):
    result = _convert_images(images=images, tmpdir_factory=tmpdir_factory)

    assert len(result.new_images) == 1

    image = result.new_images.pop()

    assert image.timepoints == 13
    assert image.depth == 12
    assert image.height == 11
    assert image.width == 10
    assert image.color_space == ColorSpace.GRAY


@pytest.mark.parametrize(
    "images,four_d",
    (
        (["image10x11x12x13-extra-stuff.mhd", "image10x11x12x13.zraw"], True),
        (["image3x4-extra-stuff.mhd", "image3x4.zraw"], False),
    ),
)
def test_staged_mhd_upload_with_additional_headers(
    tmpdir_factory, images, four_d
):
    result = _convert_images(images=images, tmpdir_factory=tmpdir_factory)

    assert len(result.new_images) == 1
    assert len(result.new_image_files) == 1

    image_file = result.new_image_files.pop()
    headers = parse_mh_header(image_file.file)

    for key in headers.keys():
        assert (key in ADDITIONAL_HEADERS) or (key in EXPECTED_HEADERS)

    for key in ADDITIONAL_HEADERS:
        assert key in headers.keys()
        if key in HEADERS_MATCHING_NUM_TIMEPOINTS:
            if four_d:
                assert len(headers[key].split(" ")) == 13
            else:
                assert len(headers[key].split(" ")) == 1

    assert "Bogus" not in headers.keys()

    original_headers = parse_mh_header(RESOURCE_PATH / images[0])
    exempt_headers = {
        "CompressedDataSize",  # Expected to change between mha/(mhd+.zraw)
        "ElementDataFile",  # Expected to change between mha/(mhd+.zraw)
        "AnatomicalOrientation",  # Isn't read but only writen based on direction
        "ElementNumberOfChannels",  # Not applicable
    }
    for key in (
        set(ADDITIONAL_HEADERS.keys()) | set(EXPECTED_HEADERS)
    ) - exempt_headers:
        if key in HEADERS_WITH_LISTING:
            for original, new in zip(
                extract_header_listing(key, original_headers),
                extract_header_listing(key, headers),
                strict=True,
            ):
                assert original == new, key
        else:
            assert original_headers[key] == headers[key], key


def test_no_convertible_file(tmpdir_factory):
    images = ["no_image", "image10x10x10.mhd", "referring_to_system_file.mhd"]

    result = _convert_images(images=images, tmpdir_factory=tmpdir_factory)

    assert len(result.new_images) == 0
    assert len(result.new_image_files) == 0
    assert len(result.file_errors) == 3


def test_errors_on_files_with_duplicate_file_names(tmpdir_factory):
    images = [
        "image10x10x10.zraw",
        "image10x10x10.mhd",
        "image10x10x10.zraw",
        "image10x10x10.mhd",
    ]

    result = _convert_images(images=images, tmpdir_factory=tmpdir_factory)

    assert len(result.new_images) == 1
    assert len(result.consumed_files) == 2


def test_mhd_file_annotation_creation(tmpdir_factory):
    images = ["image5x6x7.mhd", "image5x6x7.zraw"]

    result = _convert_images(images=images, tmpdir_factory=tmpdir_factory)

    assert len(result.new_images) == 1

    image = result.new_images.pop()
    assert image.depth == 7
    assert image.height == 6
    assert image.width == 5
    assert image.color_space == ColorSpace.GRAY


def test_subdirectory_traverse_setting(tmpdir_factory):
    input_dir = tmpdir_factory.mktemp("input")
    sub_dir = input_dir / "sub"
    sub_dir.mkdir()

    image = RESOURCE_PATH / "nifti" / "image10x11x12.nii"
    image_sub = RESOURCE_PATH / "image10x10x10.mha"

    shutil.copy(image, input_dir)
    shutil.copy(image_sub, sub_dir)
    output_dir = tmpdir_factory.mktemp("output")

    result = convert(
        input_directory=input_dir,
        output_directory=output_dir,
        recurse_subdirectories=True,
    )

    assert len(result.new_images) == 2

    output_dir = tmpdir_factory.mktemp("output2")
    result = convert(
        input_directory=input_dir,
        output_directory=output_dir,
        recurse_subdirectories=False,
    )

    assert len(result.new_images) == 1



import re
from pathlib import Path
from typing import Any

import SimpleITK

from panimg.exceptions import ValidationError

METAIO_IMAGE_TYPES = {
    "MET_NONE": None,
    "MET_ASCII_CHAR": None,
    "MET_CHAR": SimpleITK.sitkInt8,
    "MET_UCHAR": SimpleITK.sitkUInt8,
    "MET_SHORT": SimpleITK.sitkInt16,
    "MET_USHORT": SimpleITK.sitkUInt16,
    "MET_INT": SimpleITK.sitkInt32,
    "MET_UINT": SimpleITK.sitkUInt32,
    "MET_LONG": SimpleITK.sitkInt64,
    "MET_ULONG": SimpleITK.sitkUInt64,
    "MET_LONG_LONG": None,
    "MET_ULONG_LONG": None,
    "MET_FLOAT": SimpleITK.sitkFloat32,
    "MET_DOUBLE": SimpleITK.sitkFloat64,
    "MET_STRING": None,
    "MET_CHAR_ARRAY": SimpleITK.sitkVectorInt8,
    "MET_UCHAR_ARRAY": SimpleITK.sitkVectorUInt8,
    "MET_SHORT_ARRAY": SimpleITK.sitkVectorInt16,
    "MET_USHORT_ARRAY": SimpleITK.sitkVectorUInt16,
    "MET_INT_ARRAY": SimpleITK.sitkVectorInt32,
    "MET_UINT_ARRAY": SimpleITK.sitkVectorUInt32,
    "MET_LONG_ARRAY": SimpleITK.sitkVectorInt64,
    "MET_ULONG_ARRAY": SimpleITK.sitkVectorUInt64,
    "MET_LONG_LONG_ARRAY": None,
    "MET_ULONG_LONG_ARRAY": None,
    "MET_FLOAT_ARRAY": SimpleITK.sitkVectorFloat32,
    "MET_DOUBLE_ARRAY": SimpleITK.sitkVectorFloat64,
    "MET_FLOAT_MATRIX": None,
    "MET_OTHER": None,
}


FLOAT_REGEX = r"[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?"
FLOAT_MATCH_REGEXP = re.compile(rf"^{FLOAT_REGEX}$")
FLOAT_LIST_MATCH_REGEXP = re.compile(rf"^({FLOAT_REGEX})(\s{FLOAT_REGEX})*$")
FLOAT_ARRAY_MATCH_REGEXP = re.compile(
    rf"^\[({FLOAT_REGEX},\s?)*{FLOAT_REGEX}]$"
)
FLOAT_OR_FLOAT_ARRAY_MATCH_REGEX = re.compile(
    rf"({FLOAT_MATCH_REGEXP.pattern})|({FLOAT_ARRAY_MATCH_REGEXP.pattern})"
)
CONTENT_TIMES_LIST_MATCH_REGEXP = re.compile(
    r"^((2[0-3]|[0-1]\d)[0-5]\d[0-5]\d(\.\d\d\d)?)"
    r"(\s(2[0-3]|[0-1]\d)[0-5]\d[0-5]\d(\.\d\d\d)?)*$"
)

LENGTH_LIMIT_MATCH_REGEXP = re.compile(r"^.{0,128}$")

ADDITIONAL_HEADERS = {
    "Laterality": LENGTH_LIMIT_MATCH_REGEXP,
    "SliceThickness": FLOAT_MATCH_REGEXP,
    "Exposures": FLOAT_LIST_MATCH_REGEXP,
    "ContentTimes": CONTENT_TIMES_LIST_MATCH_REGEXP,
    "WindowCenter": FLOAT_OR_FLOAT_ARRAY_MATCH_REGEX,
    "WindowWidth": FLOAT_OR_FLOAT_ARRAY_MATCH_REGEX,
    "SmallestImagePixelValue": FLOAT_MATCH_REGEXP,
    "LargestImagePixelValue": FLOAT_MATCH_REGEXP,
    "t0": FLOAT_MATCH_REGEXP,
    "t1": FLOAT_MATCH_REGEXP,
}

HEADERS_MATCHING_NUM_TIMEPOINTS: list[str] = ["Exposures", "ContentTimes"]

HEADERS_MATCHING_WINDOW_SETTINGS: list[str] = ["WindowCenter", "WindowWidth"]

HEADERS_WITH_LISTING: list[str] = [
    "TransformMatrix",
    "Offset",
    "CenterOfRotation",
    "ElementSpacing",
]

EXPECTED_HEADERS: list[str] = [
    "ObjectType",
    "NDims",
    "BinaryData",
    "BinaryDataByteOrderMSB",
    "CompressedData",
    "CompressedDataSize",
    "TransformMatrix",
    "Offset",
    "CenterOfRotation",
    "AnatomicalOrientation",
    "ElementSpacing",
    "ElementNumberOfChannels",
    "DimSize",
    "ElementType",
    "ElementDataFile",
]


def parse_mh_header(file: Path) -> dict[str, str]:
    """
    Attempts to parse the headers of an mhd file.

    This function must be secure to safeguard against any untrusted uploaded
    file.

    Parameters
    ----------
    filename

    Returns
    -------
        The extracted header from the mhd file as key value pairs.

    Raises
    ------
    ValidationError
        Raised when the file contains problems making it impossible to
        read.
    """

    # attempt to limit number of read headers to prevent overflow attacks
    read_line_limit = 10000

    result: dict[str, str] = {}
    with file.open("rb") as f:
        lines = True
        while lines:
            read_line_limit -= 1
            if read_line_limit < 0:
                raise ValidationError("Files contains too many header lines")

            bin_line = f.readline(10000)

            if not bin_line:
                lines = False
                continue

            if len(bin_line) >= 10000:
                raise ValidationError("Line length is too long")

            try:
                line = bin_line.decode("utf-8")
            except UnicodeDecodeError as e:
                raise ValidationError("Header contains invalid UTF-8") from e
            else:
                result.update(extract_key_value_pairs(line))

            if "ElementDataFile" in result:
                break  # last parsed header...

    return result


def extract_key_value_pairs(line: str) -> dict[str, str]:
    line = line.rstrip("\n\r")
    if line.strip() and "=" in line:
        key, value = line.split("=", 1)
        return {key.strip(): value.strip()}
    else:
        return {}


def extract_header_listing(
    property: str, headers: dict[str, str], dtype: type = float
) -> list[Any]:
    return [dtype(e) for e in headers[property].strip().split(" ")]


def resolve_mh_data_file_path(
    headers: dict[str, str], is_mha: bool, mhd_file: Path
) -> Path:
    if is_mha:
        data_file_path = mhd_file
    else:
        data_file_path = (
            mhd_file.resolve().parent / Path(headers["ElementDataFile"]).name
        )
    if not data_file_path.exists():
        raise OSError("cannot find data file")
    return data_file_path


def validate_and_clean_additional_mh_headers(
    reader: SimpleITK.ImageFileReader,
) -> dict[str, str]:
    cleaned_headers = {}
    for key in reader.GetMetaDataKeys():
        value = reader.GetMetaData(key)
        if key in EXPECTED_HEADERS:
            cleaned_headers[key] = value
        elif key in ADDITIONAL_HEADERS:
            match_pattern = ADDITIONAL_HEADERS[key]
            if not re.match(match_pattern, value):
                raise ValidationError(
                    f"Value {value!r} for field {key!r} does not match "
                    f"pattern {match_pattern.pattern!r}"
                )
            if key in HEADERS_MATCHING_NUM_TIMEPOINTS:
                timepoints = (
                    reader.GetSize()[3] if reader.GetDimension() == 4 else 1
                )
                validate_list_data_matches_num_timepoints(
                    key=key, value=value, expected_timepoints=timepoints
                )
            if key in HEADERS_MATCHING_WINDOW_SETTINGS:
                validate_center_matches_width_setting(
                    key=key, value=value, reader=reader
                )
            cleaned_headers[key] = value
    return cleaned_headers


def validate_center_matches_width_setting(
    key: str, value: str, reader: SimpleITK.ImageFileReader
):
    window_keys = ("WindowWidth", "WindowCenter")
    if key not in window_keys:
        return
    if not re.match(FLOAT_ARRAY_MATCH_REGEXP, value):
        return

    counter_key = window_keys[0] if key == window_keys[1] else window_keys[1]
    if not reader.HasMetaDataKey(counter_key):
        raise ValidationError(
            f"Headers {key!r} and {counter_key!r} should both be present "
            f"but {counter_key!r} is missing"
        )

    counter_value = reader.GetMetaData(counter_key)
    if not re.match(FLOAT_ARRAY_MATCH_REGEXP, counter_value):
        raise ValidationError(
            f"Header {key!r} is of a different format than {counter_key!r}"
        )
    if len(value.split(",")) != len(counter_value.split(",")):
        raise ValidationError(
            f"Headers {key!r} and {counter_key!r} should "
            f"contain an equal number of values"
        )


def validate_list_data_matches_num_timepoints(
    key: str, value: str, expected_timepoints: int
):
    num_timepoints = len(value.split(" "))
    if num_timepoints != expected_timepoints:
        raise ValidationError(
            f"Found {num_timepoints} values for {key!r}, "
            f"but expected {expected_timepoints} (1/timepoint)"
        )


def add_additional_mh_headers_to_sitk_image(
    sitk_image: SimpleITK.Image, cleaned_headers: dict[str, str]
):
    for header in ADDITIONAL_HEADERS:
        if header in cleaned_headers:
            value = cleaned_headers[header]
            if isinstance(value, (list, tuple)):
                value = " ".join([str(v) for v in value])
            else:
                value = str(value)
            sitk_image.SetMetaData(header, value)


def load_sitk_image(
    file: Path, imageio: str = "MetaImageIO"
) -> SimpleITK.Image:
    # Read and validate only the header first
    reader = SimpleITK.ImageFileReader()
    reader.SetImageIO(imageio)
    reader.SetFileName(str(file.absolute()))
    reader.ReadImageInformation()

    headers = validate_and_clean_additional_mh_headers(reader=reader)

    if reader.GetNumberOfComponents() > 4:
        raise ValidationError(
            "Images with more than 4 channels not supported. "
            "For 4D data please use the 4th dimension instead."
        )

    # Header has been validated, read the pixel data
    if reader.GetDimension() in (2, 3, 4):
        sitk_image = reader.Execute()
    else:
        raise NotImplementedError(
            "SimpleITK images with more than 4 dimensions are not supported"
        )

    # Remove all headers, add back cleaned versions of supported additional headers
    for key in sitk_image.GetMetaDataKeys():
        sitk_image.EraseMetaData(key)
    add_additional_mh_headers_to_sitk_image(
        sitk_image=sitk_image, cleaned_headers=headers
    )

    return sitk_image



import re
from collections import defaultdict
from collections.abc import Iterator
from pathlib import Path
from typing import DefaultDict

from panimg.exceptions import UnconsumedFilesException, ValidationError
from panimg.image_builders.metaio_utils import load_sitk_image
from panimg.models import SimpleITKImage

MAGIC_REGEX = re.compile("NRRD([0-9]{4})")
DATAFILE_REGEX = re.compile(r"data\s?file:", flags=re.IGNORECASE)

# Arbitrary maximum lengths to prevent overflow attacks
MAX_HEADER_LINES = 10000
MAX_HEADER_LINE_LENGTH = 10000


class InvalidNrrdFileError(Exception):
    def __init__(self, message: str = "Not a NRRD image file"):
        super().__init__(message)


def verify_single_file_nrrd(file: Path) -> bool:
    """
    Reads a file and raises an InvalidNrrdFileError if that file is not a NRRD file.
    Returns True if the file is a valid NRRD file without external data file (data
    embedded in the same file). Returns False if the file is valid NRRD file but
    has one or more external data files (detached header mode).
    """
    with file.open("rb") as fp:
        # Read first 8 characters to check whether this is a NRRD file
        try:
            preamble = fp.read(8).decode("ASCII")
        except UnicodeDecodeError as e:
            raise InvalidNrrdFileError from e

        match = MAGIC_REGEX.fullmatch(preamble)
        if match is None:
            raise InvalidNrrdFileError

        # Check whether we support this version of the NRRD format
        nrrd_version = int(match.group(1))
        if nrrd_version < 1 or nrrd_version > 5:
            raise InvalidNrrdFileError("File format not supported")

        # Discard rest of preamble line
        fp.readline(MAX_HEADER_LINE_LENGTH)

        # Read rest of the header, searching for external data files
        for _ in range(MAX_HEADER_LINES):
            try:
                line = (
                    fp.readline(MAX_HEADER_LINE_LENGTH).decode("ASCII").strip()
                )
            except UnicodeDecodeError:
                break  # header cannot contain non-ASCII content

            if len(line) == 0:
                break  # empty line marks the end of the file/header

            if DATAFILE_REGEX.match(line):
                return False
        else:
            raise InvalidNrrdFileError("Header too long")

    return True


def format_error(message: str) -> str:
    return f"NRRD image builder: {message}"


def image_builder_nrrd(*, files: set[Path]) -> Iterator[SimpleITKImage]:
    """
    Constructs image objects from files in NRRD format (nrrd)

    The file format also supports detached headers, where header and data are
    stored in separate files (typically the file extension of the header is nhdr).
    This format is currently not supported since determining which files were
    consumed is complicated.

    Parameters
    ----------
    files
        Path to images that were uploaded during an upload session.

    Returns
    -------
    An `ImageBuilder` object consisting of:
     - a list of filenames for all files consumed by the image builder
     - a list of detected images
     - a list files associated with the detected images
     - path->error message map describing what is wrong with a given file
    """
    file_errors: DefaultDict[Path, list[str]] = defaultdict(list)

    for file in files:
        try:
            if verify_single_file_nrrd(file):
                img = load_sitk_image(file=file, imageio="NrrdImageIO")
            else:
                raise InvalidNrrdFileError(
                    "NRRD files with detached headers are not supported"
                )
        except (
            InvalidNrrdFileError,
            ValidationError,
            NotImplementedError,
        ) as e:
            file_errors[file].append(format_error(str(e)))
        except RuntimeError:
            file_errors[file].append(
                format_error("Not a valid NRRD image file")
            )
        else:
            yield SimpleITKImage(
                image=img,
                name=file.name,
                consumed_files={file},
                spacing_valid=True,
            )

    if file_errors:
        raise UnconsumedFilesException(file_errors=file_errors)



from pathlib import Path


class ValidationError(Exception):
    pass


class UnconsumedFilesException(Exception):  # noqa: N818
    """
    Raised on completion of an image builder and there are unconsumed files.
    Contains a dictionary with a map of the errors encountered when loading
    the unconsumed file.
    """

    def __init__(  # noqa: B042
        self, *args, file_errors: dict[Path, list[str]]
    ):
        super().__init__(*args)
        self.file_errors = file_errors



from enum import Enum

from panimg.post_processors.tiff_to_dzi import tiff_to_dzi

# DEFAULT_POST_PROCESSORS are used directly by Grand Challenge
# DO NOT CHANGE THEM without considering the impact there.
DEFAULT_POST_PROCESSORS = [tiff_to_dzi]


class PostProcessorOptions(str, Enum):
    DZI = "DZI"


POST_PROCESSOR_OPTIONS_TO_IMPLEMENTATION = {
    PostProcessorOptions.DZI: tiff_to_dzi,
}

__all__ = [
    "tiff_to_dzi",
    "DEFAULT_POST_PROCESSORS",
    "PostProcessorOptions",
    "POST_PROCESSOR_OPTIONS_TO_IMPLEMENTATION",
]



import shutil
from uuid import uuid4

from click.testing import CliRunner
from pydantic import TypeAdapter

from panimg.cli import convert_cli, post_process_cli
from panimg.models import PanImgResult, PostProcessorResult
from tests import RESOURCE_PATH


def test_convert_cli_no_files(tmp_path) -> None:
    input_dir = tmp_path / "input"
    output_dir = tmp_path / "output"

    input_dir.mkdir()

    runner = CliRunner()
    cli_result = runner.invoke(
        convert_cli,
        ["--input-dir", str(input_dir), "--output-dir", str(output_dir)],
    )
    assert cli_result.stdout == (
        '{"new_images":[],"new_image_files":[],'
        '"consumed_files":[],"file_errors":{}}\n'
    )
    assert cli_result.exit_code == 0


def test_convert_cli_with_post_processing_and_dicom(tmp_path) -> None:
    input_dir = tmp_path / "input"
    output_dir = tmp_path / "output"

    input_dir.mkdir()

    # 3 new files, all supported
    shutil.copy(str(RESOURCE_PATH / "image10x10x10.mha"), str(input_dir))
    shutil.copy(str(RESOURCE_PATH / "valid_tiff.tif"), str(input_dir))
    shutil.copy(str(RESOURCE_PATH / "dicom_2d" / "cxr.dcm"), str(input_dir))

    runner = CliRunner()
    cli_result = runner.invoke(
        convert_cli,
        ["--input-dir", str(input_dir), "--output-dir", str(output_dir)],
    )
    assert cli_result.exit_code == 0

    assert len(cli_result.stdout.splitlines()) == 1

    panimg_result: PanImgResult = TypeAdapter(PanImgResult).validate_json(
        cli_result.stdout.splitlines()[-1]
    )

    assert {im.name for im in panimg_result.new_images} == {
        "1.2.392.200036.9125.0.199302241758.16-0",
        "valid_tiff.tif",
        "image10x10x10.mha",
    }
    assert len(panimg_result.new_image_files) == 4
    assert {im.image_type for im in panimg_result.new_image_files} == {
        "DZI",
        "TIFF",
        "MHD",
    }


def test_convert_cli_with_verbosity(tmp_path) -> None:
    input_dir = tmp_path / "input"
    output_dir = tmp_path / "output"

    input_dir.mkdir()

    shutil.copy(str(RESOURCE_PATH / "image10x10x10.mha"), str(input_dir))

    runner = CliRunner()
    cli_result = runner.invoke(
        convert_cli,
        [
            "--input-dir",
            str(input_dir),
            "--output-dir",
            str(output_dir),
            "-vvv",
            "--image-builder",
            "MHD",
            "--no-post-processing",
        ],
    )
    assert cli_result.exit_code == 0

    assert cli_result.stdout.splitlines()[0].startswith("Converting ")

    # Last line should be the JSON result
    panimg_result: PanImgResult = TypeAdapter(PanImgResult).validate_json(
        cli_result.stdout.splitlines()[-1]
    )

    assert {im.name for im in panimg_result.new_images} == {
        "image10x10x10.mha",
    }
    assert len(panimg_result.new_image_files) == 1
    assert {im.image_type for im in panimg_result.new_image_files} == {
        "MHD",
    }


def test_convert_cli_with_post_processing_no_dicom(tmp_path) -> None:
    input_dir = tmp_path / "input"
    output_dir = tmp_path / "output"

    input_dir.mkdir()

    # 3 new files, DICOM should be skipped
    shutil.copy(str(RESOURCE_PATH / "image10x10x10.mha"), str(input_dir))
    shutil.copy(str(RESOURCE_PATH / "valid_tiff.tif"), str(input_dir))
    shutil.copy(str(RESOURCE_PATH / "dicom_2d" / "cxr.dcm"), str(input_dir))

    runner = CliRunner()
    cli_result = runner.invoke(
        convert_cli,
        [
            "--input-dir",
            str(input_dir),
            "--output-dir",
            str(output_dir),
            "--image-builder",
            "MHD",
            "--image-builder",
            "TIFF",
        ],
    )
    assert cli_result.exit_code == 0

    panimg_result: PanImgResult = TypeAdapter(PanImgResult).validate_json(
        cli_result.stdout.splitlines()[-1]
    )

    assert {im.name for im in panimg_result.new_images} == {
        "valid_tiff.tif",
        "image10x10x10.mha",
    }
    assert len(panimg_result.new_image_files) == 3
    assert {im.image_type for im in panimg_result.new_image_files} == {
        "DZI",
        "TIFF",
        "MHD",
    }


def test_convert_cli_with_defined_post_processing_no_dicom(tmp_path) -> None:
    input_dir = tmp_path / "input"
    output_dir = tmp_path / "output"

    input_dir.mkdir()

    # 3 new files, DICOM should be skipped
    shutil.copy(str(RESOURCE_PATH / "image10x10x10.mha"), str(input_dir))
    shutil.copy(str(RESOURCE_PATH / "valid_tiff.tif"), str(input_dir))
    shutil.copy(str(RESOURCE_PATH / "dicom_2d" / "cxr.dcm"), str(input_dir))

    runner = CliRunner()
    cli_result = runner.invoke(
        convert_cli,
        [
            "--input-dir",
            str(input_dir),
            "--output-dir",
            str(output_dir),
            "--image-builder",
            "TIFF",
            "--post-processor",
            "DZI",
        ],
    )
    assert cli_result.exit_code == 0

    panimg_result: PanImgResult = TypeAdapter(PanImgResult).validate_json(
        cli_result.stdout.splitlines()[-1]
    )

    assert {im.name for im in panimg_result.new_images} == {
        "valid_tiff.tif",
    }
    assert len(panimg_result.new_image_files) == 2
    assert {im.image_type for im in panimg_result.new_image_files} == {
        "DZI",
        "TIFF",
    }


def test_convert_cli_no_post_processing_no_dicom(tmp_path) -> None:
    input_dir = tmp_path / "input"
    output_dir = tmp_path / "output"

    input_dir.mkdir()

    # 3 new files, DICOM and post processing should be skipped
    shutil.copy(str(RESOURCE_PATH / "image10x10x10.mha"), str(input_dir))
    shutil.copy(str(RESOURCE_PATH / "valid_tiff.tif"), str(input_dir))
    shutil.copy(str(RESOURCE_PATH / "dicom_2d" / "cxr.dcm"), str(input_dir))

    runner = CliRunner()
    cli_result = runner.invoke(
        convert_cli,
        [
            "--input-dir",
            str(input_dir),
            "--output-dir",
            str(output_dir),
            "--image-builder",
            "MHD",
            "--image-builder",
            "TIFF",
            "--no-post-processing",
        ],
    )
    assert cli_result.exit_code == 0

    panimg_result: PanImgResult = TypeAdapter(PanImgResult).validate_json(
        cli_result.stdout.splitlines()[-1]
    )

    assert {im.name for im in panimg_result.new_images} == {
        "valid_tiff.tif",
        "image10x10x10.mha",
    }
    assert len(panimg_result.new_image_files) == 2
    assert {im.image_type for im in panimg_result.new_image_files} == {
        "TIFF",
        "MHD",
    }


def test_convert_cli_no_post_processing_and_dicom(tmp_path) -> None:
    input_dir = tmp_path / "input"
    output_dir = tmp_path / "output"

    input_dir.mkdir()

    # 3 new files, all supported
    shutil.copy(str(RESOURCE_PATH / "image10x10x10.mha"), str(input_dir))
    shutil.copy(str(RESOURCE_PATH / "valid_tiff.tif"), str(input_dir))
    shutil.copy(str(RESOURCE_PATH / "dicom_2d" / "cxr.dcm"), str(input_dir))

    runner = CliRunner()
    cli_result = runner.invoke(
        convert_cli,
        [
            "--input-dir",
            str(input_dir),
            "--output-dir",
            str(output_dir),
            "--no-post-processing",
        ],
    )
    assert cli_result.exit_code == 0

    panimg_result: PanImgResult = TypeAdapter(PanImgResult).validate_json(
        cli_result.stdout.splitlines()[-1]
    )

    assert {im.name for im in panimg_result.new_images} == {
        "1.2.392.200036.9125.0.199302241758.16-0",
        "valid_tiff.tif",
        "image10x10x10.mha",
    }
    assert len(panimg_result.new_image_files) == 3
    assert {im.image_type for im in panimg_result.new_image_files} == {
        "TIFF",
        "MHD",
    }


def test_post_process_cli_no_dzi(tmp_path) -> None:
    target_file = tmp_path / "input.bin"
    shutil.copy(RESOURCE_PATH / "image10x10x10.mha", target_file)

    runner = CliRunner()
    cli_result = runner.invoke(
        post_process_cli,
        [
            "--image-id",
            str(uuid4()),
            "--image-type",
            "MHD",
            "--input-file",
            target_file,
        ],
    )
    assert cli_result.stdout == '{"new_image_files":[]}\n'
    assert cli_result.exit_code == 0

    post_processor_result: PostProcessorResult = TypeAdapter(
        PostProcessorResult
    ).validate_json(cli_result.stdout.splitlines()[-1])

    assert post_processor_result.new_image_files == set()


def test_post_process_cli_dzi(tmp_path) -> None:
    target_file = tmp_path / "input.bin"
    shutil.copy(RESOURCE_PATH / "valid_tiff.tif", target_file)

    runner = CliRunner()
    cli_result = runner.invoke(
        post_process_cli,
        [
            "--image-id",
            str(uuid4()),
            "--image-type",
            "TIFF",
            "--input-file",
            target_file,
        ],
    )
    assert cli_result.exit_code == 0

    post_processor_result: PostProcessorResult = TypeAdapter(
        PostProcessorResult
    ).validate_json(cli_result.stdout.splitlines()[-1])

    assert len(post_processor_result.new_image_files) == 1
    assert {im.image_type for im in post_processor_result.new_image_files} == {
        "DZI"
    }


def test_post_process_cli_dzi_processor_set(tmp_path) -> None:
    target_file = tmp_path / "input.bin"
    shutil.copy(RESOURCE_PATH / "valid_tiff.tif", target_file)

    runner = CliRunner()
    cli_result = runner.invoke(
        post_process_cli,
        [
            "--image-id",
            str(uuid4()),
            "--image-type",
            "TIFF",
            "--input-file",
            target_file,
            "--post-processor",
            "DZI",
        ],
    )
    assert cli_result.exit_code == 0

    post_processor_result: PostProcessorResult = TypeAdapter(
        PostProcessorResult
    ).validate_json(cli_result.stdout.splitlines()[-1])

    assert len(post_processor_result.new_image_files) == 1
    assert {im.image_type for im in post_processor_result.new_image_files} == {
        "DZI"
    }


def test_post_process_cli_dzi_processor_set_with_verbosity(tmp_path) -> None:
    target_file = tmp_path / "input.bin"
    shutil.copy(RESOURCE_PATH / "valid_tiff.tif", target_file)

    runner = CliRunner()
    cli_result = runner.invoke(
        post_process_cli,
        [
            "--image-id",
            str(uuid4()),
            "--image-type",
            "TIFF",
            "--input-file",
            target_file,
            "--post-processor",
            "DZI",
            "-vvv",
        ],
    )
    assert cli_result.exit_code == 0

    assert cli_result.stdout.splitlines()[0].startswith("Post processing ")

    # Last line should be the JSON result
    post_processor_result: PostProcessorResult = TypeAdapter(
        PostProcessorResult
    ).validate_json(cli_result.stdout.splitlines()[-1])

    assert len(post_processor_result.new_image_files) == 1
    assert {im.image_type for im in post_processor_result.new_image_files} == {
        "DZI"
    }



from collections.abc import Iterable
from enum import Enum

from panimg.image_builders.dicom import image_builder_dicom
from panimg.image_builders.fallback import image_builder_fallback
from panimg.image_builders.metaio_mhd_mha import image_builder_mhd
from panimg.image_builders.metaio_nifti import image_builder_nifti
from panimg.image_builders.metaio_nrrd import image_builder_nrrd
from panimg.image_builders.oct import image_builder_oct
from panimg.image_builders.tiff import image_builder_tiff
from panimg.types import ImageBuilder

# DEFAULT_IMAGE_BUILDERS are used directly by Grand Challenge
# DO NOT CHANGE THEM without considering the impact there.
DEFAULT_IMAGE_BUILDERS: Iterable[ImageBuilder] = [
    image_builder_mhd,
    image_builder_nifti,
    image_builder_nrrd,
    image_builder_dicom,
    image_builder_tiff,
    image_builder_oct,
    image_builder_fallback,
]


class ImageBuilderOptions(str, Enum):
    MHD = "MHD"
    NIFTI = "NIFTI"
    NRRD = "NRRD"
    DICOM = "DICOM"
    TIFF = "TIFF"
    OCT = "OCT"
    FALLBACK = "FALLBACK"


IMAGE_BUILDER_OPTIONS_TO_IMPLEMENTATION: dict[str, ImageBuilder] = {
    ImageBuilderOptions.MHD: image_builder_mhd,
    ImageBuilderOptions.NIFTI: image_builder_nifti,
    ImageBuilderOptions.NRRD: image_builder_nrrd,
    ImageBuilderOptions.DICOM: image_builder_dicom,
    ImageBuilderOptions.TIFF: image_builder_tiff,
    ImageBuilderOptions.OCT: image_builder_oct,
    ImageBuilderOptions.FALLBACK: image_builder_fallback,
}

__all__ = [
    "image_builder_mhd",
    "image_builder_nifti",
    "image_builder_nrrd",
    "image_builder_dicom",
    "image_builder_tiff",
    "image_builder_oct",
    "image_builder_fallback",
    "DEFAULT_IMAGE_BUILDERS",
    "ImageBuilderOptions",
    "IMAGE_BUILDER_OPTIONS_TO_IMPLEMENTATION",
]



import os
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from panimg.image_builders.metaio_nrrd import image_builder_nrrd
from panimg.models import ColorSpace
from panimg.panimg import _build_files
from tests import RESOURCE_PATH


@pytest.mark.parametrize(
    "srcfile",
    (
        RESOURCE_PATH / "nrrd" / "image10x11x12.nrrd",
        RESOURCE_PATH / "nrrd" / "image10x11x12_compressed.nrrd",
    ),
)
def test_image_builder_nrrd(tmpdir_factory, srcfile: Path):
    dest = Path(tmpdir_factory.mktemp("input"))

    # Copy files into an empty temporary directory
    shutil.copy(srcfile, dest)
    files = {dest / srcfile.name}

    # Run image builder
    result = _build_files(
        builder=image_builder_nrrd,
        files=files,
        output_directory=tmpdir_factory.mktemp("output"),
    )

    # Verify results
    assert result.consumed_files == files
    assert len(result.new_images) == 1

    image = result.new_images.pop()
    assert image.color_space == ColorSpace.GRAY.value
    assert image.width == 10
    assert image.height == 11
    assert image.depth == 12
    assert image.voxel_width_mm == pytest.approx(1.0)
    assert image.voxel_height_mm == pytest.approx(2.0)
    assert image.voxel_depth_mm == pytest.approx(3.0)


@pytest.mark.parametrize(
    "srcfiles",
    (
        [
            RESOURCE_PATH / "nrrd" / "image10x11x12.nhdr",
            RESOURCE_PATH / "nrrd" / "image10x11x12.raw",
        ],
        [
            RESOURCE_PATH / "nrrd" / "image10x11x12_compressed.nhdr",
            RESOURCE_PATH / "nrrd" / "image10x11x12_compressed.raw.gz",
        ],
    ),
)
def test_image_builder_nrrd_detached_header(
    tmpdir_factory, srcfiles: list[Path]
):
    dest = Path(tmpdir_factory.mktemp("input"))

    # Copy files into an empty temporary directory
    files = set()
    for srcfile in srcfiles:
        shutil.copy(srcfile, dest)
        files.add(dest / srcfile.name)

    # Run image builder
    result = _build_files(
        builder=image_builder_nrrd,
        files=files,
        output_directory=tmpdir_factory.mktemp("output"),
    )

    # Verify results (should not read anything)
    assert result.consumed_files == set()
    assert len(result.new_images) == 0


def test_image_builder_with_unsupported_version(tmpdir):
    dest = Path(tmpdir) / "image10x11x12_unsupported_version.nrrd"
    shutil.copy(RESOURCE_PATH / "nrrd" / dest.name, dest)
    files = {Path(d[0]).joinpath(f) for d in os.walk(tmpdir) for f in d[2]}
    with TemporaryDirectory() as output:
        result = _build_files(
            builder=image_builder_nrrd, files=files, output_directory=output
        )
    assert result.consumed_files == set()
    assert len(result.new_images) == 0


def test_image_builder_with_other_file_extension(tmpdir):
    dest = Path(tmpdir) / "image10x10x10.mha"
    shutil.copy(RESOURCE_PATH / dest.name, dest)
    files = {Path(d[0]).joinpath(f) for d in os.walk(tmpdir) for f in d[2]}
    with TemporaryDirectory() as output:
        result = _build_files(
            builder=image_builder_nrrd, files=files, output_directory=output
        )
    assert result.consumed_files == set()
    assert len(result.new_images) == 0



from pathlib import Path
from tempfile import TemporaryDirectory

import pytest
from tifffile import TiffFile
from wsidicom import WsiDicom

from panimg.contrib.wsi_dcm_to_tiff.dcm_to_tiff import dcm_to_tiff
from tests import RESOURCE_PATH


@pytest.mark.parametrize(
    "src",
    (
        RESOURCE_PATH / "dicom_wsi/sparse_with_bot",
        RESOURCE_PATH / "dicom_wsi/sparse_no_bot",
        RESOURCE_PATH / "dicom_wsi/full_with_bot",
        RESOURCE_PATH / "dicom_wsi/full_no_bot",
    ),
)
def test_dcm_to_tiff(src: Path):
    with TemporaryDirectory() as output_directory:
        converted_tiff = Path(output_directory) / "output.tiff"
        dcm_to_tiff(src, converted_tiff)
        files = {f for f in src.rglob("*") if f.is_file()}
        with WsiDicom.open(files) as o_tif:
            with TiffFile(converted_tiff) as c_tif:
                assert len(o_tif.levels) == len(c_tif.pages)
                for i in range(0, len(o_tif.levels)):
                    assert (
                        o_tif.collection[i].size.height
                        == c_tif.pages[i].shape[0]
                    )
                    assert (
                        o_tif.collection[i].size.width
                        == c_tif.pages[i].shape[1]
                    )
                    assert (
                        o_tif.collection[i].tile_size.height
                        == c_tif.pages[i].tile[0]
                    )
                    assert (
                        o_tif.collection[i].tile_size.width
                        == c_tif.pages[i].tile[1]
                    )



import os
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from panimg.image_builders.metaio_nifti import image_builder_nifti
from panimg.models import ColorSpace
from panimg.panimg import _build_files
from tests import RESOURCE_PATH


@pytest.mark.parametrize(
    "src",
    (
        RESOURCE_PATH / "nifti" / "image10x11x12.nii",
        RESOURCE_PATH / "nifti" / "image10x11x12.nii.gz",
    ),
)
def test_image_builder_nifti(tmpdir_factory, src: Path):
    dest = Path(tmpdir_factory.mktemp("input"))

    shutil.copy(src, dest / src.name)

    files = {*dest.glob("*")}

    result = _build_files(
        builder=image_builder_nifti,
        files=files,
        output_directory=tmpdir_factory.mktemp("output"),
    )

    assert result.consumed_files == files
    assert len(result.new_images) == 1

    image = result.new_images.pop()
    assert image.color_space == ColorSpace.GRAY.value
    assert image.width == 10
    assert image.height == 11
    assert image.depth == 12
    assert image.voxel_width_mm == pytest.approx(1.0)
    assert image.voxel_height_mm == pytest.approx(2.0)
    assert image.voxel_depth_mm == pytest.approx(3.0)


def test_image_builder_with_other_file_extension(tmpdir):
    dest = Path(tmpdir) / "image10x10x10.mha"
    shutil.copy(RESOURCE_PATH / dest.name, dest)
    files = {Path(d[0]).joinpath(f) for d in os.walk(tmpdir) for f in d[2]}
    with TemporaryDirectory() as output:
        result = _build_files(
            builder=image_builder_nifti, files=files, output_directory=output
        )
    assert result.consumed_files == set()
    assert len(result.new_images) == 0



import logging
import shutil
from enum import Enum
from functools import cached_property
from pathlib import Path
from uuid import UUID, uuid4

import numpy as np
import SimpleITK
from pydantic import BaseModel, ConfigDict, field_validator
from pydantic.dataclasses import dataclass
from SimpleITK import GetArrayViewFromImage, Image, WriteImage

logger = logging.getLogger(__name__)

MASK_TYPE_PIXEL_IDS = [
    SimpleITK.sitkInt8,
    SimpleITK.sitkUInt8,
]


class ColorSpace(str, Enum):
    GRAY = "GRAY"
    RGB = "RGB"
    RGBA = "RGBA"
    YCBCR = "YCBCR"


ITK_COLOR_SPACE_MAP = {
    1: ColorSpace.GRAY,
    3: ColorSpace.RGB,
    4: ColorSpace.RGBA,
}


class ImageType(str, Enum):
    MHD = "MHD"
    TIFF = "TIFF"
    DZI = "DZI"


class EyeChoice(str, Enum):
    OCULUS_DEXTER = "OD"
    OCULUS_SINISTER = "OS"
    UNKNOWN = "U"
    NOT_APPLICABLE = "NA"


# NOTE: Only int8 or uint8 data types are checked for segments
# so the true maximum is 256
MAXIMUM_SEGMENTS_LENGTH = 64


@dataclass(frozen=True)
class PanImg:
    pk: UUID
    name: str
    width: int
    height: int
    depth: int | None
    voxel_width_mm: float | None
    voxel_height_mm: float | None
    voxel_depth_mm: float | None
    timepoints: int | None
    resolution_levels: int | None
    window_center: float | None
    window_width: float | None
    color_space: ColorSpace
    eye_choice: EyeChoice
    segments: frozenset[int] | None = None


@dataclass(frozen=True)
class PanImgFile:
    image_id: UUID
    image_type: ImageType
    file: Path
    directory: Path | None = None


@dataclass
class PanImgResult:
    new_images: set[PanImg]
    new_image_files: set[PanImgFile]
    consumed_files: set[Path]
    file_errors: dict[Path, list[str]]


@dataclass
class PostProcessorResult:
    new_image_files: set[PanImgFile]


class SimpleITKImage(BaseModel):
    image: Image

    name: str
    consumed_files: set[Path]

    spacing_valid: bool
    eye_choice: EyeChoice = EyeChoice.NOT_APPLICABLE

    model_config = ConfigDict(arbitrary_types_allowed=True, frozen=True)

    @property
    def width(self) -> int:
        return int(self.image.GetWidth())

    @property
    def height(self) -> int:
        return int(self.image.GetHeight())

    @property
    def depth(self) -> int | None:
        try:
            depth = int(self.image.GetDepth())
        except (RuntimeError, ValueError):
            return None

        return depth or None

    @staticmethod
    def _extract_first_float(value: str) -> float:
        if value.startswith("["):
            return float(value[1:-1].split(",")[0])
        else:
            return float(value)

    @property
    def window_center(self) -> float | None:
        try:
            return self._extract_first_float(
                self.image.GetMetaData("WindowCenter")
            )
        except (RuntimeError, ValueError):
            return None

    @property
    def window_width(self) -> float | None:
        try:
            return self._extract_first_float(
                self.image.GetMetaData("WindowWidth")
            )
        except (RuntimeError, ValueError):
            return None

    @property
    def timepoints(self) -> int | None:
        if self.image.GetDimension() == 4 and self.segments is None:
            # Only 4D files that are non-segmentations have timepoints
            return int(self.image.GetSize()[3])
        else:
            return None

    @field_validator("image")
    @classmethod
    def check_color_space(cls, image: Image):  # noqa: B902, N805
        cs = image.GetNumberOfComponentsPerPixel()
        if cs not in ITK_COLOR_SPACE_MAP:
            raise ValueError(f"Unknown color space for MetaIO image: {cs}")
        return image

    @field_validator("image")
    @classmethod
    def add_value_range_meta_data(cls, image: Image):  # noqa: B902, N805
        smallest_tag = "SmallestImagePixelValue"
        largest_tag = "LargestImagePixelValue"

        if image.HasMetaDataKey(smallest_tag) and image.HasMetaDataKey(
            largest_tag
        ):
            return image

        # Use the numpy-array route to support a larger range of data-types
        # than ITK' MinimumMaximumImageFilter (e.g. 2D uint8)
        array = GetArrayViewFromImage(image)
        image.SetMetaData(smallest_tag, str(array.min()))
        image.SetMetaData(largest_tag, str(array.max()))

        return image

    @cached_property
    def segments(self) -> frozenset[int] | None:
        if (
            self.image.GetNumberOfComponentsPerPixel() != 1
            or self.image.GetPixelIDValue() not in MASK_TYPE_PIXEL_IDS
        ):
            # Only single channel, 8 bit images should be checked.
            # Everything else is not a segmentation
            return None

        im_arr = GetArrayViewFromImage(self.image)

        if self.image.GetDimension() == 4:
            segments = set()
            n_volumes = self.image.GetSize()[3]

            for volume in range(n_volumes):
                # Calculate the segments for each volume for memory efficiency
                volume_segments = np.unique(im_arr[volume, :, :, :])
                segments.update({*volume_segments})

                if not segments.issubset({0, 1}):
                    # 4D Segmentations must only have values 0 and 1
                    # as the 4th dimension encodes the overlay type
                    return None

            # Use 1-indexing for each segmentation
            segments = {idx + 1 for idx in range(n_volumes)}
        else:
            segments = np.unique(im_arr)

        if len(segments) <= MAXIMUM_SEGMENTS_LENGTH:
            return frozenset(segments)
        else:
            return None

    @property
    def color_space(self) -> ColorSpace:
        return ITK_COLOR_SPACE_MAP[self.image.GetNumberOfComponentsPerPixel()]

    @property
    def voxel_width_mm(self) -> float | None:
        if self.spacing_valid:
            return float(self.image.GetSpacing()[0])
        else:
            return None

    @property
    def voxel_height_mm(self) -> float | None:
        if self.spacing_valid:
            return float(self.image.GetSpacing()[1])
        else:
            return None

    @property
    def voxel_depth_mm(self) -> float | None:
        if self.spacing_valid:
            try:
                return float(self.image.GetSpacing()[2])
            except IndexError:
                return None
        else:
            return None

    def save(self, output_directory: Path) -> tuple[PanImg, set[PanImgFile]]:
        pk = uuid4()

        work_dir = Path(output_directory) / self.name
        work_dir.mkdir()

        new_image = PanImg(
            pk=pk,
            name=self.name,
            width=self.width,
            height=self.height,
            depth=self.depth,
            window_center=self.window_center,
            window_width=self.window_width,
            timepoints=self.timepoints,
            resolution_levels=None,
            color_space=self.color_space,
            voxel_width_mm=self.voxel_width_mm,
            voxel_height_mm=self.voxel_height_mm,
            voxel_depth_mm=self.voxel_depth_mm,
            eye_choice=self.eye_choice,
            segments=self.segments,
        )

        WriteImage(
            image=self.image,
            fileName=str(work_dir.absolute() / f"{pk}.mha"),
            useCompression=True,
        )

        new_files = set()
        for file in work_dir.iterdir():
            new_file = PanImgFile(
                image_id=pk, image_type=ImageType.MHD, file=file
            )
            new_files.add(new_file)

        return new_image, new_files


class TIFFImage(BaseModel):
    file: Path

    name: str
    consumed_files: set[Path]

    width: int
    height: int
    voxel_width_mm: float
    voxel_height_mm: float
    resolution_levels: int
    color_space: ColorSpace
    eye_choice: EyeChoice = EyeChoice.NOT_APPLICABLE
    segments: frozenset[int] | None = None

    model_config = ConfigDict(frozen=True)

    def save(self, output_directory: Path) -> tuple[PanImg, set[PanImgFile]]:
        pk = uuid4()

        output_file = output_directory / self.name / f"{pk}{self.file.suffix}"
        output_file.parent.mkdir()

        new_image = PanImg(
            pk=pk,
            name=self.name,
            width=self.width,
            height=self.height,
            depth=1,
            resolution_levels=self.resolution_levels,
            color_space=self.color_space,
            voxel_width_mm=self.voxel_width_mm,
            voxel_height_mm=self.voxel_height_mm,
            voxel_depth_mm=None,
            timepoints=None,
            window_center=None,
            window_width=None,
            eye_choice=self.eye_choice,
            segments=self.segments,
        )

        shutil.copy(src=self.file, dst=output_file)

        new_files = {
            PanImgFile(
                image_id=pk,
                image_type=ImageType.TIFF,
                file=output_file.absolute(),
            )
        }

        return new_image, new_files



import math
from pathlib import Path

import numpy as np
from tifffile import TiffWriter

try:
    from isyntax import ISyntax
except ImportError:
    _has_isyntax = False
else:
    _has_isyntax = True


def is_isyntax(dir_path):
    try:
        path = Path(dir_path)
        files = list(path.iterdir())
        with ISyntax.open(files):
            return True
    except Exception:
        return False


def isyntax_to_tiff(input_path, output_path):
    """Convert iSyntax WSI to TIFF format.

    This function converts an iSyntax WSI file to a TIFF file using the pyisyntax
    library.

    Parameters
    ----------
    input_path: str
        Path to the iSyntax WSI file.
    output_path: str
        Path to the output TIFF file."""
    if not _has_isyntax:
        raise ImportError("pyisyntax not installed")
    path = Path(input_path)
    image = ISyntax.open(path)
    wsi = image.wsi

    with TiffWriter(output_path, bigtiff=True) as tif:

        def tiler(getter, level_idx, cols, rows):
            for row in range(rows):
                for col in range(cols):
                    im_rgb = np.asarray(
                        getter(
                            col * image.tile_width,
                            row * image.tile_width,
                            image.tile_width,
                            image.tile_height,
                            level_idx,
                        )
                    )[..., :3]
                    yield im_rgb

        for level_idx, level in enumerate(wsi.levels):
            col = int(math.ceil(level.width / image.tile_width))
            row = int(math.ceil(level.height / image.tile_height))

            shape = (level.height, level.width, 3)

            level_tiler = tiler(image.read_region, level_idx, col, row)

            resolution = (
                int(1e4 / level.mpp_x),
                int(1e4 / level.mpp_y),
            )

            sub_filetype = 1 if level_idx != 0 else None

            tif.write(
                level_tiler,
                dtype="uint8",
                shape=shape,
                tile=(image.tile_height, image.tile_width),
                photometric="rgb",
                compression="jpeg",
                subsampling=(1, 1),
                resolution=resolution,
                resolutionunit="CENTIMETER",
                description="Converted from iSyntax",
                subfiletype=sub_filetype,
            )

    image.close()



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any

import pytest

from panimg.image_builders import image_builder_dicom, image_builder_tiff
from panimg.models import ColorSpace, EyeChoice
from panimg.panimg import _build_files
from tests import RESOURCE_PATH

DEFAULT_EXPECTED_IMAGE_PROPERTIES = {
    "width": 154,
    "height": 290,
    "voxel_width_mm": 0.125,
    "voxel_height_mm": 0.125,
    "resolution_levels": 1,
    "depth": 1,
    "voxel_depth_mm": None,
    "timepoints": None,
    "window_center": None,
    "window_width": None,
    "color_space": ColorSpace.YCBCR,
    "eye_choice": EyeChoice.NOT_APPLICABLE,
    "segments": None,
}


@pytest.mark.parametrize(
    "src,expected_image,errors",
    (
        (
            RESOURCE_PATH / "dicom_wsi/sparse_with_bot",
            DEFAULT_EXPECTED_IMAGE_PROPERTIES,
            0,
        ),
        (
            RESOURCE_PATH / "dicom_wsi/sparse_no_bot",
            DEFAULT_EXPECTED_IMAGE_PROPERTIES,
            0,
        ),
        (
            RESOURCE_PATH / "dicom_wsi/full_with_bot",
            DEFAULT_EXPECTED_IMAGE_PROPERTIES,
            0,
        ),
        (
            RESOURCE_PATH / "dicom_wsi/full_no_bot",
            DEFAULT_EXPECTED_IMAGE_PROPERTIES,
            0,
        ),
        (RESOURCE_PATH / "dicom_wsi/non_wsi_dcm", {}, 1),
        (RESOURCE_PATH / "isyntax_wsi", {}, 0),
    ),
)
def test_image_builder_wsi(
    src: Path, expected_image: dict[str, Any], errors: int
):
    files = {f for f in src.rglob("*") if f.is_file()}
    if not files and src.name == "isyntax_wsi":
        pytest.xfail(
            reason="iSyntax resource is not available. "
            "To run this test, download the file using --download-files."
        )
    with TemporaryDirectory() as output:
        result = _build_files(
            builder=image_builder_tiff,
            files=files,
            output_directory=Path(output),
        )

    assert len(result.file_errors) == errors
    if errors == 0:
        assert result.consumed_files == files
        assert len(result.new_images) == 1
        for res in result.new_images:
            for k, v in expected_image.items():
                assert getattr(res, k) == v


def test_dicom_wsi_fails_for_dicom_builder(tmpdir):
    src = RESOURCE_PATH / "dicom_wsi/sparse_with_bot"
    files = {f for f in src.rglob("*") if f.is_file()}
    result = _build_files(
        builder=image_builder_dicom, files=files, output_directory=tmpdir
    )

    assert len(result.new_image_files) == 0
    assert len(result.file_errors) == 1



from collections import defaultdict
from collections.abc import Iterator
from pathlib import Path
from typing import DefaultDict

from panimg.exceptions import UnconsumedFilesException, ValidationError
from panimg.image_builders.metaio_utils import load_sitk_image
from panimg.models import SimpleITKImage


def format_error(message: str) -> str:
    return f"NifTI image builder: {message}"


def image_builder_nifti(*, files: set[Path]) -> Iterator[SimpleITKImage]:
    """
    Constructs image objects from files in NifTI format (nii/nii.gz)

    Parameters
    ----------
    files
        Path to images that were uploaded during an upload session.

    Returns
    -------
    An `ImageBuilder` object consisting of:
     - a list of filenames for all files consumed by the image builder
     - a list of detected images
     - a list files associated with the detected images
     - path->error message map describing what is wrong with a given file
    """
    file_errors: DefaultDict[Path, list[str]] = defaultdict(list)

    for file in files:
        if not (file.name.endswith(".nii") or file.name.endswith(".nii.gz")):
            file_errors[file].append(format_error("Not a NifTI image file"))
            continue

        try:
            img = load_sitk_image(file=file, imageio="NiftiImageIO")
        except (ValidationError, NotImplementedError) as e:
            file_errors[file].append(format_error(str(e)))
        except RuntimeError:
            file_errors[file].append(
                format_error("Not a valid NifTI image file")
            )
        else:
            yield SimpleITKImage(
                image=img,
                name=file.name,
                consumed_files={file},
                spacing_valid=True,
            )

    if file_errors:
        raise UnconsumedFilesException(file_errors=file_errors)



import os
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from panimg.image_builders.oct import format_error, image_builder_oct
from panimg.panimg import _build_files
from tests import RESOURCE_PATH


@pytest.mark.parametrize(
    "src,expected_fundus_properties,expected_oct_properties",
    (
        (
            # Minimized .fda OCT file was created by taking example OCT file at
            # biobank.ndph.ox.ac.uk/showcase/showcase/examples/eg_oct_fda.fda
            # and downsizing:
            #   - OCT volume (@IMG_JPEG) to 512x650x1 J2C encoded,
            #   - Fundus (@IMG_FUNDUS) to 2x3x1 J2C encoded
            RESOURCE_PATH / "oct/fda_minimized.fda",
            {
                "width": 2,
                "height": 3,
                "depth": None,
                "voxel_width_mm": None,
                "voxel_height_mm": None,
                "voxel_depth_mm": None,
                "eye_choice": "U",
            },
            {
                "width": 512,
                "height": 650,
                "depth": 1,
                "voxel_width_mm": 0.01171875,
                "voxel_height_mm": 0.0035,
                "voxel_depth_mm": 6,
                "eye_choice": "U",
            },
        ),
        (
            # Minimized .fds OCT file was created by taking example OCT file at
            # biobank.ndph.ox.ac.uk/showcase/showcase/examples/eg_oct_fds.fds
            # and downsizing:
            #   - OCT volume (@IMG_SCAN_03) to 2x3x4 16bit voxels,
            #   - OBS (@IMG_OBS) scan to 2x3x1 24bit voxels,
            #   - MOT_COMP (@IMG_MOT_COMP_03) to 2x3x4 16bit voxels,
            #   - TRC (@IMG_TRC_02) to 2x3x4 24bit voxels
            RESOURCE_PATH / "oct/fds_minimized.fds",
            {
                "width": 2,
                "height": 3,
                "depth": None,
                "voxel_width_mm": None,
                "voxel_height_mm": None,
                "voxel_depth_mm": None,
                "eye_choice": "U",
            },
            {
                "width": 2,
                "height": 3,
                "depth": 4,
                "voxel_width_mm": 3.0,
                "voxel_height_mm": 0.0035,
                "voxel_depth_mm": 1.5,
                "eye_choice": "U",
            },
        ),
        (
            # Minimized .e2e OCT file was created by taking example OCT file from RUMC
            # and downsizing:
            #   - OCT volume (chunk type = 1073741824, chunk ind = 1) to 512x496x1
            #   - Fundus image (chunk type = 1073741824, chunk ind = 0) to 2x3x1
            #   - deleting all MDbData chunks other than the OCT slice, the fundus
            #     image and the laterality information (chunk type = 11)
            RESOURCE_PATH / "oct/e2e_minimized.E2E",
            {
                "width": 2,
                "height": 3,
                "depth": None,
                "voxel_width_mm": None,
                "voxel_height_mm": None,
                "voxel_depth_mm": None,
                "eye_choice": "U",
            },
            {
                "width": 512,
                "height": 496,
                "depth": 1,
                "voxel_width_mm": 0.01171875,
                "voxel_height_mm": 0.0039,
                "voxel_depth_mm": 4.5,
                "eye_choice": "U",
            },
        ),
    ),
)
def test_image_builder_oct(
    tmpdir, src, expected_fundus_properties, expected_oct_properties
):
    dest = Path(tmpdir) / src.name
    shutil.copy(str(src), str(dest))
    files = {Path(d[0]).joinpath(f) for d in os.walk(tmpdir) for f in d[2]}
    with TemporaryDirectory() as output:
        result = _build_files(
            builder=image_builder_oct, files=files, output_directory=output
        )

    assert result.consumed_files == {dest}
    assert len(result.new_images) == 1
    for res in result.new_images:
        expected_values = expected_oct_properties
        if "fundus" in res.name:
            # expected_values = expected_fundus_properties
            continue  # Skip fundus_images for now

        for k, v in expected_values.items():
            assert getattr(res, k) == v


def test_image_builder_oct_corrupt_file(tmpdir):
    src = RESOURCE_PATH / "oct/corrupt.fds"
    dest = Path(tmpdir) / src.name
    shutil.copy(str(src), str(dest))

    files = {Path(d[0]).joinpath(f) for d in os.walk(tmpdir) for f in d[2]}
    with TemporaryDirectory() as output:
        result = _build_files(
            builder=image_builder_oct, files=files, output_directory=output
        )

    assert result.file_errors == {
        dest: [
            format_error(
                "Not a valid OCT file " "(supported formats: .fds,.fda,.e2e)"
            )
        ]
    }
    assert result.consumed_files == set()



import os
from collections import defaultdict
from pathlib import Path
from unittest import mock

import numpy as np
import pydicom
import pytest
import SimpleITK

from panimg.image_builders.dicom import (
    PixelValueInverter,
    _find_valid_dicom_files,
    format_error,
    get_dicom_headers_by_study,
    image_builder_dicom,
)
from panimg.image_builders.metaio_utils import parse_mh_header
from panimg.panimg import _build_files
from tests import RESOURCE_PATH

DICOM_DIR = RESOURCE_PATH / "dicom_4d"


def test_get_dicom_headers_by_study():
    files = [Path(d[0]).joinpath(f) for d in os.walk(DICOM_DIR) for f in d[2]]
    studies = get_dicom_headers_by_study(files, defaultdict(list))
    assert len(studies) == 1
    for key in studies:
        assert {x["file"] for x in studies[key]["headers"]} == {
            DICOM_DIR / f"{x}.dcm" for x in range(1, 77)
        }

    for root, _, files in os.walk(RESOURCE_PATH):
        files = [Path(root).joinpath(f) for f in files]
        break

    studies = get_dicom_headers_by_study(files, defaultdict(list))
    assert len(studies) == 0


def test_validate_dicom_files():
    files = {Path(d[0]).joinpath(f) for d in os.walk(DICOM_DIR) for f in d[2]}
    studies = _find_valid_dicom_files(files, defaultdict(list))
    assert len(studies) == 1
    for study in studies:
        headers = study.headers
        assert study.n_time == 19
        assert study.n_slices == 4
    with mock.patch(
        "panimg.image_builders.dicom.get_dicom_headers_by_study",
        return_value={
            "foo": {"headers": headers[1:], "name": "StudyInstanceUID-0"}
        },
    ):
        errors = defaultdict(list)
        studies = _find_valid_dicom_files(files, errors)
        assert len(studies) == 0
        for header in headers[1:]:
            assert errors[header["file"]] == [
                format_error("Number of slices per time point differs")
            ]


def test_image_builder_dicom_single_slice(tmpdir):
    files = {DICOM_DIR / "1.dcm"}
    result = _build_files(
        builder=image_builder_dicom, files=files, output_directory=tmpdir
    )
    assert result.consumed_files == files
    assert len(result.new_images) == 1

    image = result.new_images.pop()
    assert image.depth == 1
    assert image.voxel_depth_mm == pytest.approx(1.0)


def test_image_builder_dicom_2d(tmpdir):
    files = {RESOURCE_PATH / "dicom_2d" / "cxr.dcm"}
    result = _build_files(
        builder=image_builder_dicom, files=files, output_directory=tmpdir
    )
    assert result.consumed_files == files
    assert len(result.new_images) == 1

    image = result.new_images.pop()
    assert image.width == 440
    assert image.height == 440
    assert image.depth == 1

    assert image.voxel_width_mm == pytest.approx(0.1)
    assert image.voxel_height_mm == pytest.approx(0.1)
    assert image.voxel_depth_mm == pytest.approx(1.0)

    sitk_image = SimpleITK.ReadImage(str(result.new_image_files.pop().file))
    assert sitk_image.GetDimension() == 3
    assert np.allclose(sitk_image.GetOrigin(), (0, 0, 0))
    assert np.allclose(sitk_image.GetDirection(), (1, 0, 0, 0, 1, 0, 0, 0, 1))
    assert sitk_image.GetPixelID() == SimpleITK.sitkUInt16

    # Photometric interpretation is MONOCHROME1 so pixel values have been
    # altered (inverted), including the window level but not the width
    array = SimpleITK.GetArrayViewFromImage(sitk_image)
    assert np.count_nonzero(array == 979) == 6
    assert np.count_nonzero(array == 0) == 4
    assert sitk_image.GetMetaData("WindowCenter") == "429"
    assert sitk_image.GetMetaData("WindowWidth") == "1024"


def test_image_builder_dicom_4d(tmpdir):
    files = {Path(d[0]).joinpath(f) for d in os.walk(DICOM_DIR) for f in d[2]}
    result = _build_files(
        builder=image_builder_dicom, files=files, output_directory=tmpdir
    )
    assert result.consumed_files == {
        Path(DICOM_DIR).joinpath(f"{x}.dcm") for x in range(1, 77)
    }

    assert len(result.new_images) == 1

    image = result.new_images.pop()
    assert image.timepoints == 19
    assert image.depth == 4
    assert image.height == 2
    assert image.width == 3

    assert len(result.new_image_files) == 1
    mha_file_obj = [
        x for x in result.new_image_files if x.file.suffix == ".mha"
    ][0]

    headers = parse_mh_header(mha_file_obj.file)

    direction = headers["TransformMatrix"].split()
    origin = headers["Offset"].split()
    spacing = headers["ElementSpacing"].split()
    exposures = headers["Exposures"].split()
    content_times = headers["ContentTimes"].split()

    assert len(exposures) == 19
    assert exposures == [str(x) for x in range(100, 2000, 100)]
    assert len(content_times) == 19
    assert content_times == [str(x) for x in range(214501, 214520)]

    dcm_ref = pydicom.dcmread(str(DICOM_DIR / "1.dcm"))
    assert np.array_equal(
        np.array(list(map(float, direction))).reshape((4, 4)), np.eye(4)
    )
    assert np.allclose(
        list(map(float, spacing))[:2],
        list(map(float, list(dcm_ref.PixelSpacing))),
    )
    assert np.allclose(
        list(map(float, origin)),
        list(map(float, dcm_ref.ImagePositionPatient)) + [0.0],
    )


def test_image_builder_dicom_enhanced():
    # Load reference image from individual slices
    slices = set((RESOURCE_PATH / "dicom_enhanced").glob("slice*.dcm"))
    ref_images = list(image_builder_dicom(files=slices))
    assert len(ref_images) == 1
    ref_image = ref_images.pop().image

    # Load the same image but now stored as enhanced DICOM volume
    volume = {RESOURCE_PATH / "dicom_enhanced" / "volume.dcm"}
    images = list(image_builder_dicom(files=volume))
    assert len(images) == 1
    image = images.pop().image

    assert image == ref_image


def test_image_builder_dicom_4d_enhanced():
    # Load reference image from individual slices
    slices = set((RESOURCE_PATH / "dicom_4d").glob("*.dcm"))
    ref_images = list(image_builder_dicom(files=slices))
    assert len(ref_images) == 1
    ref_image = ref_images.pop().image

    # Load the same image but now stored as enhanced DICOM volume
    volumes = set((RESOURCE_PATH / "dicom_4d_enhanced").glob("*.dcm"))
    images = list(image_builder_dicom(files=volumes))
    assert len(images) == 1
    image = images.pop().image

    for i in range(ref_image.GetSize()[0]):
        # Compare in 3D because 4D comparisons are not implemented in SITK
        assert image[i, :, :, :] == ref_image[i, :, :, :]


@pytest.mark.parametrize(
    "directory,element_type",
    [
        ("dicom_4d", "MET_SHORT"),
        ("dicom_intercept", "MET_FLOAT"),
        ("dicom_slope", "MET_FLOAT"),
    ],
)
def test_dicom_rescaling(directory, element_type, tmpdir):
    """
    2.dcm in dicom_intercept and dicom_slope has been modified to add a
    small intercept (0.01) or slope (1.001) respectively.
    """
    files = [
        Path(d[0]).joinpath(f)
        for d in os.walk(RESOURCE_PATH / directory)
        for f in d[2]
    ]
    result = _build_files(
        builder=image_builder_dicom, files=files, output_directory=tmpdir
    )

    assert len(result.new_image_files) == 1
    mha_file_obj = [
        x for x in result.new_image_files if x.file.suffix == ".mha"
    ][0]

    headers = parse_mh_header(mha_file_obj.file)
    assert headers["ElementType"] == element_type


@pytest.mark.parametrize(
    "files,center,center_ob,width,width_ob",
    [
        (
            {
                Path(d[0]).joinpath(f)
                for d in os.walk(RESOURCE_PATH / "dicom_4d")
                for f in d[2]
            },
            "30",
            30.0,
            "200",
            200.0,
        ),
        (
            [
                RESOURCE_PATH / "dicom_window_level" / "1.dcm",
                RESOURCE_PATH / "dicom_window_level" / "2.dcm",
            ],
            "[10.5, 20.5, 30.5]",
            10.5,
            "[1.5, 2.5, 3.5]",
            1.5,
        ),
    ],
)
def test_dicom_window_level(tmpdir, files, center, center_ob, width, width_ob):
    result = _build_files(
        builder=image_builder_dicom, files=files, output_directory=tmpdir
    )

    assert len(result.new_image_files) == 1
    mha_file_obj = [
        x for x in result.new_image_files if x.file.suffix == ".mha"
    ][0]

    headers = parse_mh_header(mha_file_obj.file)
    assert headers["WindowCenter"] == center
    assert headers["WindowWidth"] == width

    assert len(result.new_images) == 1
    image_obj = result.new_images.pop()
    assert image_obj.window_center == center_ob
    assert image_obj.window_width == width_ob


@pytest.mark.parametrize(
    "array_as_list,expected",
    [
        ([0], [0]),
        ([-1, 0, 1], [1, 0, -1]),
        ([0, 1, 3], [3, 2, 0]),
        ([5, 6, 7], [7, 6, 5]),
        ([0.5, 0.6, 0.7], [0.7, 0.6, 0.5]),
        (
            [
                [0, 1, 3],
                [3, 0, 1],
            ],
            [
                [3, 2, 0],
                [0, 3, 2],
            ],
        ),
    ],
)
def test_dicom_photometric_interpretation_inversion(array_as_list, expected):
    array = np.array(array_as_list)
    inverter = PixelValueInverter(array)
    np.testing.assert_equal(inverter.invert(array), np.array(expected))



from pathlib import Path

import pytest

from panimg import image_builders
from panimg.image_builders.metaio_utils import load_sitk_image
from panimg.models import SimpleITKImage
from panimg.panimg import _build_files
from tests import RESOURCE_PATH


@pytest.mark.parametrize(
    "src_image_name,smallest_pixel_value,largest_pixel_value",
    [
        (  # Regular case
            "image_min10_max10.mha",
            -10,
            10,
        ),
        (  # Predefined ranges in input file
            "image3x4-extra-stuff.mhd",
            -100,
            100,
        ),
        (  # 3D image
            "image10x10x10.mha",
            8.341192e-05,
            0.999185,
        ),
        (  # 4D image
            "image10x11x12x13.mha",
            2,
            2,
        ),
    ],
)
def test_sitk_image_value_range(
    src_image_name,
    smallest_pixel_value,
    largest_pixel_value,
):
    src = RESOURCE_PATH / src_image_name
    sitk_image = load_sitk_image(src)
    result = SimpleITKImage(
        image=sitk_image,
        name=src.name,
        consumed_files={src},
        spacing_valid=True,
    )

    for value, tag in [
        (smallest_pixel_value, "SmallestImagePixelValue"),
        (largest_pixel_value, "LargestImagePixelValue"),
    ]:
        if value is None:
            assert not result.image.HasMetaDataKey(tag)
        else:
            assert float(result.image.GetMetaData(tag)) == pytest.approx(value)


@pytest.mark.parametrize(
    "src_image,builder,segments",
    [
        (
            "segments/-10_10_Int8.mha",
            image_builders.image_builder_mhd,
            frozenset(range(-10, 10)),
        ),
        (
            "segments/0_10_UInt8.mha",
            image_builders.image_builder_mhd,
            frozenset(range(10)),
        ),
        (
            "segments/4D_1_1_1_5_UInt8.mha",
            image_builders.image_builder_mhd,
            frozenset({1, 2, 3, 4, 5}),
        ),
        (
            # Contains non-zero or ones as values
            "segments/4D_1_1_1_5_UInt8_threes.mha",
            image_builders.image_builder_mhd,
            None,
        ),
        (
            # Datatype is not sitkUInt8 or sitkInt8
            "image_min10_max10.mha",
            image_builders.image_builder_mhd,
            None,
        ),
        (  # Too many values
            "dicom_2d/cxr.dcm",
            image_builders.image_builder_dicom,
            None,
        ),
        (  # Image type is vector of ints
            "test_rgb.png",
            image_builders.image_builder_fallback,
            None,
        ),
        (  # Tiffs are always None
            "valid_tiff.tif",
            image_builders.image_builder_tiff,
            None,
        ),
    ],
)
def test_segments(
    src_image,
    builder,
    segments,
    tmpdir_factory,
):
    files = {RESOURCE_PATH / src_image}
    output_dir = Path(tmpdir_factory.mktemp("output"))
    result = _build_files(
        builder=builder, files=files, output_directory=output_dir
    )
    assert result.consumed_files == files
    assert len(result.new_images) == 1

    image = result.new_images.pop()
    assert image.segments == segments


def test_invalid_4d(tmp_path_factory):
    file = RESOURCE_PATH / "channels" / "5_1_1_1_invalid.mha"
    result = _build_files(
        builder=image_builders.image_builder_mhd,
        files={file},
        output_directory=tmp_path_factory.mktemp("output"),
    )

    assert result.consumed_files == set()
    assert len(result.new_images) == 0
    assert result.file_errors == {
        file: [
            (
                "Mhd image builder: Images with more than 4 channels not supported. "
                "For 4D data please use the 4th dimension instead."
            )
        ]
    }


def test_4d_segmentation_none_timepoints(tmp_path_factory):
    result = _build_files(
        builder=image_builders.image_builder_mhd,
        files={RESOURCE_PATH / "segments" / "4D_1_1_1_5_UInt8.mha"},
        output_directory=tmp_path_factory.mktemp("output"),
    )

    new_image = result.new_images.pop()

    assert new_image.segments == frozenset({1, 2, 3, 4, 5})
    assert new_image.timepoints is None


def test_model_strips_headers(tmpdir):
    src = RESOURCE_PATH / "image3x4-extra-stuff.mhd"
    old_image = load_sitk_image(src)
    old_image.SetMetaData("PatientID", "remove_me")

    result = SimpleITKImage(
        image=old_image,
        name=src.name,
        consumed_files={src},
        spacing_valid=True,
    )
    _, new_files = result.save(output_directory=tmpdir)

    new_image = load_sitk_image(new_files.pop().file)

    removed_keys = {*old_image.GetMetaDataKeys()} - {
        *new_image.GetMetaDataKeys()
    }

    assert removed_keys == {"PatientID"}
    assert {*new_image.GetMetaDataKeys()} == {
        "ContentTimes",
        "Exposures",
        "LargestImagePixelValue",
        "Laterality",
        "SliceThickness",
        "SmallestImagePixelValue",
        "WindowCenter",
        "WindowWidth",
        "t0",
        "t1",
    }
